import getpass
import logging
import os
import re
import sys

from localconfig import LocalConfig
from workspace.commands import AbstractCommand
from workspace.scm import repo_check, repo_path, commit_logs, extract_commit_msgs, is_git_repo
from workspace.utils import silent_run


log = logging.getLogger(__name__)
new_version = None  # Doesn't work if it is in bump_version
PUBLISH_VERSION_PREFIX = 'Publish version '
IGNORE_CHANGE_RE = re.compile('^(?:Update changelog|Fix tests?)\s*$', flags=re.IGNORECASE)


class Publish(AbstractCommand):
  """
    Bumps version in setup.py (defaults to patch), writes out changelog, builds a source distribution, and uploads with twine.

    :param bool minor: Perform a minor publish by bumping the minor version
    :param bool major: Perform a major publish by bumping the major version
  """
  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [
      cls.make_args('--minor', action='store_true', help=docs['minor']),
      cls.make_args('--major', action='store_true', help=docs['major'])
    ]

  def run(self):
    if self.minor and self.major:
      log.error('--minor and --major are mutually exclusive, please use only one.')
      return

    repo_check()

    pypirc = LocalConfig('~/.pypirc')
    if not pypirc.pypi or not pypirc.pypi.repository:
      log.error('Please add repository / username to [pypi] section in ~/.pypirc')
      sys.exit(1)

    if not pypirc.pypi.username:
      pypirc.pypi.username = getpass.getuser('PyPI Username: ')
      if not pypirc.pypi.username:
        sys.exit()

    if not pypirc.pypi.password:
      pypirc.pypi.password = getpass.getpass('PyPI Password: ')
      if not pypirc.pypi.password:
        sys.exit()

    self.commander.run('update')

    changes = self.changes_since_last_publish()

    if not changes:
      log.info('There are no changes since last publish')
      sys.exit(0)

    silent_run('rm -rf dist/*', shell=True, cwd=repo_path())

    new_version = self.bump_version()
    self.update_changelog(new_version, changes, self.minor or self.major)

    self.commander.run('commit', msg=PUBLISH_VERSION_PREFIX + new_version, push=True)

    log.info('Building source distribution')
    silent_run('python setup.py sdist', cwd=repo_path())

    log.info('Uploading')

    silent_run('twine upload -u "{username}" -p "{password}" dist/*'.format(**dict(list(pypirc.pypi))), shell=True, cwd=repo_path())

  def changes_since_last_publish(self):
    commit_msgs = extract_commit_msgs(commit_logs(limit=100, repo=repo_path()), is_git_repo())
    changes = []

    for msg in commit_msgs:
      if msg.startswith(PUBLISH_VERSION_PREFIX):
        break
      if len(msg) < 7 or IGNORE_CHANGE_RE.match(msg):
        continue
      changes.append(msg)

    return changes

  def update_changelog(self, new_version, changes, skip_title_change=False):
    docs_dir = os.path.join(repo_path(), 'docs')
    if not os.path.isdir(docs_dir):
      os.makedirs(docs_dir)

    changelog_file = os.path.join(docs_dir, 'CHANGELOG.rst')
    existing_changes = os.path.exists(changelog_file) and open(changelog_file).read()
    major_title = '=' * 80
    minor_title = '-' * 80

    with open(changelog_file, 'w') as fp:
      fp.write('Version %s' % new_version + '\n')
      fp.write(major_title + '\n\n')

      for change in changes:
        fp.write('* %s\n' % (change.replace('\n', '\n  ')))

      if existing_changes:
        fp.write('\n')
        if not skip_title_change:
          existing_changes = existing_changes.replace(major_title, minor_title, 1)
        fp.write(existing_changes)

  def bump_version(self):
    """
      Bump the version (defaults to patch) in setup.py
    """
    setup_file = os.path.join(repo_path(), 'setup.py')

    if not os.path.exists(setup_file):
      log.error(setup_file + ' does not exist.')
      sys.exit(1)

    def replace_version(match):
      global new_version

      version_parts = match.group(2).split('.')
      i = 0 if self.major else (1 if self.minor else 2)

      while len(version_parts) < i + 1:
        version_parts.append(0)

      for j in range(i+1, len(version_parts)):
        version_parts[j] = '0'

      version_parts[i] = str(int(version_parts[i]) + 1)
      new_version = '.'.join(version_parts)

      return 'version=' + match.group(1) + new_version + match.group(1)

    content = re.sub('version\s*=\s*([\'"])(.*)[\'"]', replace_version, open(setup_file).read())

    with open(setup_file, 'w') as fp:
      fp.write(content)

    if not new_version:
      log.error('Failed to find "version=" in setup.py to bump version')
      sys.exit(1)

    return new_version



import logging

from bumper import BumperDriver

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.commands.review import Review
from workspace.config import config
from workspace.scm import repo_check, is_git_repo, product_name, current_branch


log = logging.getLogger(__name__)


class Bump(AbstractCommand):
  """
    Bump dependency versions in requirements.txt, pinned.txt, or any specified file.

    :param str names: Only bump dependencies that match the name.
                      Name can be a product group name defined in workspace.cfg.
                      To bump to a specific version instead of latest, append version to name
                      (i.e. requests==1.2.3 or 'requests>=1.2.3'). When > or < is used, be sure to quote.
    :param int test: Run tests. Results will be posted to ReviewBoard if --rb is used.
    :param bool rb: Create or update existing RB after commit. Existing RB is looked up in .git/config.
    :param bool push: Wait for 'Ship It' from RB (unless --skip-rb is used) and push the bump (git only)
    :param bool add: Add the `names` to the requirements file if they don't exist.
    :param str msg: Summary commit message
    :param str/list file: Requirement file to bump. Defaults to requirements.txt or pinned.txt
                          that are set by bump.requirement_files in workspace.cfg.
    :param dict bumper_models: List of classes that implements :class:`bumper.cars.AbstractBumper`
                               Defaults to :class:`bumper.cars.RequirementsBumper`
    :param bool force: Force a bump even when certain bump requirements are not met.
    :param bool dry_run: Perform a dry run by printing out the changes only without making changes.
    :param dict kwargs: Additional args from argparse
  """

  def __init__(self, *args, **kwargs):
    kwargs.setdefault('show_filter', True)
    super(Bump, self).__init__(*args, **kwargs)

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return ([
        cls.make_args('names', nargs='*', help=docs['names']),
        cls.make_args('--add', action='store_true', help=docs['add']),
        cls.make_args('--force', action='store_true', help=docs['force']),
        cls.make_args('-m', '--msg', help=docs['msg']),
        cls.make_args('--file', help=docs['file']),
        cls.make_args('-n', '--dry-run', action='store_true', help=docs['dry_run'])
      ], [
        cls.make_args('-t', '--test', action='count', help=docs['test']),
        cls.make_args('-r', '--rb', action='store_true', help=docs['rb']),
        cls.make_args('-p', '--push', action='store_true', help=docs['push']),
      ])

  def run(self):
    """
      :return: Tuple with 3 elements: A map of file to bump message, commit message, and list of :class:`Bump`
    """
    repo_check()

    self.commander.run('update')

    if not self.msg:
      config.commit.auto_branch_from_commit_words = 1

    if not self.names:
      self.names = []

    filter_requirements = expand_product_groups(self.names)

    if self.show_filter and filter_requirements:
      log.info('Only bumping: %s', ' '.join(filter_requirements))

    if isinstance(self.file, list):
      requirement_files = self.file
    elif self.file:
      requirement_files = [self.file]
    else:
      requirement_files = config.bump.requirement_files.strip().split()

    bumper = BumperDriver(requirement_files, bumper_models=self.bumper_models, full_throttle=self.force, detail=True, test_drive=self.dry_run)
    messages, bumps = bumper.bump(filter_requirements, required=self.add, show_summary=not is_git_repo())
    commit_msg = None

    try:
      if messages:
        summary_msgs = []
        detail_msgs = []
        for m in sorted(messages.values()):
          splits = m.split('\n', 1)
          summary_msgs.append(splits[0])
          if len(splits) == 2:
            detail_msgs.append(splits[1])

        commit_msg = '\n\n'.join(summary_msgs + detail_msgs)

        if self.msg:
          commit_msg = self.msg + '\n\n' + commit_msg

        if not self.dry_run and is_git_repo():
          self.commander.run('commit', msg=commit_msg, files=messages.keys())

    except Exception:
      bumper.reverse()
      raise

    if bumps:
      tests = {}

      if self.test:
        log.info('Running tests')
        tests[product_name()] = self.commander.run('test', return_output=self.rb, test_dependents=self.test > 1)

      if not self.dry_run:
        if self.rb and commit_msg and self.commander.command('review') != Review:
          reviewer_groups = set()
          reviewers = set()

          for lib in bumps.keys():
            g, r = self.commander.command('review').reviewers_for_product(lib)
            if g:
              reviewer_groups.update(g)
            if r:
              reviewers.update(r)
            log.debug('Reviewers for %s: %s %s', lib, g, r)

          self.commander.run('review', publish=self.push, files=messages.keys(), description=commit_msg, test=tests,
                             skip_prereview=True, reviewer_groups=reviewer_groups, reviewers=reviewers)

        if self.push and is_git_repo():
          branch = current_branch()
          if self.rb:
            self.commander.run('wait', review=True, in_background=True)
          self.commander.run('push', branch=branch)

    return messages, commit_msg, bumps



import logging

from workspace.commands import AbstractCommand
from workspace.scm import (checkout_branch, remove_branch, git_repo_check, current_branch, update_repo,
                           push_repo, merge_branch, update_branch)


log = logging.getLogger(__name__)


class Push(AbstractCommand):
  """
    Push changes for branch

    :param bool push: The branch to push. Defaults to current branch.
  """
  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [cls.make_args('branch', nargs='?', help=docs['push'])]

  def run(self):

    git_repo_check()

    if not self.branch:
      self.branch = current_branch()

    log.info('Pushing %s', self.branch)

    checkout_branch('master')

    update_repo()

    if self.branch != 'master':
      checkout_branch(self.branch)
      update_branch()  # Failed rebase can be continued easily than failed merge

      checkout_branch('master')
      merge_branch(self.branch)

    push_repo()

    if self.branch != 'master':
      remove_branch(self.branch)



import logging
import os
import shutil
from time import time

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.config import config
from workspace.scm import workspace_path, product_name, repos, stat_repo, all_branches, is_git_repo
from workspace.utils import silent_run

log = logging.getLogger(__name__)


class Clean(AbstractCommand):
  """ Clean workspace by removing build, dist, and .pyc files """

  def run(self):

    path = workspace_path()
    log.info('Cleaning %s', path)

    if config.clean.remove_products_older_than_days or config.clean.remove_all_products_except:
      keep_time = 0
      keep_products = []

      if config.clean.remove_all_products_except:
        log.info('Removing all products except: %s', config.clean.remove_all_products_except)
        keep_products = expand_product_groups(config.clean.remove_all_products_except.split())

      if config.clean.remove_products_older_than_days:
        log.info('Removing products older than %s days', config.clean.remove_products_older_than_days)
        keep_time = time() - config.clean.remove_products_older_than_days * 86400

      removed_products = []

      for repo in repos(path):
        name = product_name(repo)
        modified_time = os.stat(repo).st_mtime
        if name not in keep_products or keep_time and modified_time < keep_time:
          status = stat_repo(repo, return_output=True)
          if not status or 'nothing to commit' in status and 'working directory clean' in status and is_git_repo(repo) and len(all_branches(repo)) <= 1:
            shutil.rmtree(repo)
            removed_products.append(name)
          else:
            log.info('  - Skipping "%s" as it has changes that may not be committed', name)

      if removed_products:
        log.info('Removed %s', ', '.join(removed_products))

    log.info('Removing build/dist folders')
    silent_run("rm -rf */build */dist */docs/_build", cwd=path, shell=True)

    log.info('Removing *.pyc files')
    silent_run("find . -type d \( -path '*/.tox' -o -path '*/mppy-*' \) -prune -o -name *.pyc -exec rm {} \;", cwd=path, shell=True)



from glob import glob
import logging
import os
import re
import subprocess

from localconfig import LocalConfig

from workspace.config import product_groups

log = logging.getLogger(__name__)


class ToxIni(LocalConfig):
  """ Represents tox.ini """

  VAR_RE = re.compile('{\[(.+)](.+)}')

  def __init__(self, repo=None, tox_ini=None):
    """
    :param str repo: The repo to load tox*.ini from.
    :param str tox_ini: Path to tox ini file. Defaults to tox*.ini in repo root.
    """
    self.repo = repo
    self.path = tox_ini or self.path_for(repo)
    super(ToxIni, self).__init__(self.path)

  @classmethod
  def path_for(cls, repo):
    """
    :param str repo: Repo to get tox*.ini for
    :return: Path to tox*.ini
    :raise IOError: if there is no tox*.ini found
    """

    tox_inis = glob(os.path.join(repo, 'tox*.ini'))

    if not tox_inis:
      raise IOError('No tox.ini found in %s. Please run "wst setup --product" first to setup tox.' % repo)

    elif len(tox_inis) > 1:
      log.warn('More than one ini files found - will use first one: %s', ', '.join(tox_inis))

    return tox_inis[0]

  @property
  def envlist(self):
    return [e.strip() for e in self.tox.envlist.split(',') if e]

  def envsection(self, env):
    return 'testenv:%s' % env

  @property
  def workdir(self):
    return os.path.join(self.repo, self.get('tox', 'toxworkdir', '.tox'))

  def envdir(self, env):
    envsection = self.envsection(env)
    if envsection not in self:
      log.debug('Using default envdir and commands as %s section is not defined in %s', envsection, self.path)

    return self.expand_vars(self.get(envsection, 'envdir', os.path.join(self.repo, '.tox', env)).replace('{toxworkdir}', self.workdir))

  def bindir(self, env, script=None):
    dir = os.path.join(self.envdir(env), 'bin')
    if script:
      dir = os.path.join(dir, script)
    return dir

  def commands(self, env):
    envsection = self.envsection(env)
    commands = self.get(envsection, 'commands', self.get('testenv', 'commands', 'py.test {env:PYTESTARGS:}'))
    return filter(None, self.expand_vars(commands).split('\n'))

  def expand_vars(self, value):
    if '{' in value:
      value = self.VAR_RE.sub(lambda m: self.get(m.group(1), m.group(2)), value)
    return value


class ProductPager(object):
  """ Pager to show contents from multiple products (repos) """
  MAX_TERMINAL_ROWS = 25

  def __init__(self, optional=False):
      """ If optional is True, then pager is only used if the # of lines of the first write exceeds `self.MAX_TERMINAL_ROWS` lines. """
      self.pager = None
      self.optional = optional

  def write(self, product, output, branch=None):
    if not self.pager:
      if not self.optional or len(output.split('\n')) > self.MAX_TERMINAL_ROWS:
        self.pager = create_pager('^\[.*]')

    if self.pager:
      self.pager.stdin.write('[ ' + product + ' ]\n')
      if branch and branch != 'master':
        self.pager.stdin.write('# On branch %s\n' % branch)
      self.pager.stdin.write(output.strip() + '\n\n')
    else:
      if branch and branch != 'master':
        print '# On branch %s' % branch
      print output

  def close_and_wait(self):
    if self.pager:
      self.pager.stdin.close()
      self.pager.wait()


def create_pager(highlight_text=None):
  """ Returns a pipe to PAGER or "less" """
  pager_cmd = os.environ.get('PAGER')

  if not pager_cmd:
    pager_cmd = ['less']
    if highlight_text:
      pager_cmd.extend(['-p', highlight_text])

  pager = subprocess.Popen(pager_cmd, stdin=subprocess.PIPE)

  return pager


def expand_product_groups(names):
  """ Expand product groups found in the given list of names to produce a sorted list of unique names. """
  unique_names = set(names)
  exclude_names = set()

  for name in list(unique_names):
    if name.startswith('-'):
      unique_names.remove(name)
      exclude_names.update(expand_product_groups([name.lstrip('-')]))

  for group, names in product_groups().items():
    if group in unique_names:
      unique_names.remove(group)
      unique_names.update(expand_product_groups(names))

  return sorted(list(unique_names - exclude_names))



"""
# Default config. Change these in your personal ~/.config/workspace.cfg ::

  ###########################################################################################################
  # Define product groups to take action upon (such as wst checkout, develop, or bump)
  ###########################################################################################################
  [product_groups]
  mzheng-repos =
    git@github.com:maxzheng/workspace-tools.git
    git@github.com:maxzheng/clicast.git
    git@github.com:maxzheng/localconfig.git
    git@github.com:maxzheng/remoteconfig.git
  mzheng = workspace-tools clicast localconfig remoteconfig


  ###########################################################################################################
  # Settings for bump command
  ###########################################################################################################
  [bump]

  # List of requirement files to check / bump dependencies in
  requirement_files = requirements.txt pinned.txt


  ###########################################################################################################
  # Settings for checkout command
  ###########################################################################################################
  [checkout]

  # Check out SVN repo using git-svn and clone the specified # of commits. Set to 0 to disable git-svn clone.
  # Higher the number, the longer it takes to clone.
  use_gitsvn_to_clone_svn_commits = 10

  # API used to find git repo for single word checkout (i.e. wst checkout workspace-tools)
  # It should accept a ?q=singleWord param
  search_api_url = https://api.github.com/search/repositories

  # URL to use when checking out a user repo reference (i.e. wst checkout maxzheng/workspace-tools)
  user_repo_url = git@github.com:%s.git


  ###########################################################################################################
  # Settings for clean command
  ###########################################################################################################
  [clean]

  # Remove products that have not been modified since given days ago
  remove_products_older_than_days =

  # Remove all products except for these ones (product or group)
  remove_all_products_except =

  ###########################################################################################################
  # Settings for commit command
  ###########################################################################################################
  [commit]

  # Automatically create branch based on the first number of commit words. Set to 0 to turn off.
  auto_branch_from_commit_words = 2
"""

import logging
import os

from remoteconfig import RemoteConfig


CONFIG_FILE = 'workspace.cfg'
USER_CONFIG_FILE = os.path.join('~', '.config', CONFIG_FILE)

log = logging.getLogger()
config = RemoteConfig(USER_CONFIG_FILE, cache_duration=60)
config.read(__doc__.replace('\n  ', '\n'))


def product_groups():
  """ Returns a dict with product group name mapped to products """
  return dict((group, names.split()) for group, names in config.product_groups)



import argparse
import logging
import pkg_resources
import sys
import textwrap

from workspace.commands.bump import Bump
from workspace.commands.checkout import Checkout
from workspace.commands.clean import Clean
from workspace.commands.commit import Commit
from workspace.commands.test import Test
from workspace.commands.diff import Diff
from workspace.commands.log import Log
from workspace.commands.publish import Publish
from workspace.commands.push import Push
from workspace.commands.review import Review
from workspace.commands.update import Update
from workspace.commands.status import Status
from workspace.commands.setup import Setup
from workspace.commands.wait import Wait
from workspace.utils import log_exception


log = logging.getLogger(__name__)


class Commander(object):
  """
    Tools to simplify workspace / scm management when working with multiple repositories.

    It helps you do more with less work by seamlessly integrating all workspace tooling into one where
    you can simply run one command instead of many native commands individually to do common tasks.

    To get started
    ---------------
    * Optionally setup workspace environment/shortcuts, run "wst setup -h" for options.
        - "wst setup -a" is recommended. :)
    * To checkout a product, run: wst checkout <git or svn repository url> [<url2> ...]
        - Product that use svn is checked out with git-svn
    * All commands are named appropriately for what they do, but see its --help for additional info.
    * For more info, read the docs at http://workspace-tools.readthedocs.org
  """

  @classmethod
  def commands(cls):
    """
      Map of command name to command classes.
      Override commands to replace any command name with another class to customize the command.
    """
    cs = [Bump, Checkout, Clean, Commit, Diff, Log, Publish, Push, Review, Setup, Status, Test, Update, Wait]
    return dict((c.name(), c) for c in cs)

  @classmethod
  def command(cls, name):
    """ Get command class for name """
    return cls.commands().get(name)

  @classmethod
  def main(cls):
    logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
    cls()._run()

  def _run(self):
    """
      Sets up logging, parser, and creates the necessary command sequences to run, and runs
      the command given by the user.
    """
    self.setup_parsers()

    args, extra_args = self.parser.parse_known_args()

    if args.command not in [c.name() for c in self.commands().values() if 'extra_args' in c.docs()[1]] and extra_args:
      log.error('Unrecognized arguments: %s', ' '.join(extra_args))
      sys.exit(1)

    if args.debug:
      logging.root.setLevel(logging.DEBUG)

    with log_exception(exit=True, stack=args.debug):
      args_dict = args.__dict__
      args_dict['extra_args'] = extra_args
      return self.run(args.command, **args_dict)

  def run(self, name=None, **kwargs):
    """
      Run the command by name with given args.

      :param str name: Name of command to run. If not given, this calls self._run()
      :param kwargs: Args to pass to the command constructor
    """
    if not name:
      return self._run()

    if name in self.commands():
      kwargs['commander'] = self
      return self.command(name)(**kwargs).run()
    else:
      log.error('Command "%s" is not registered. Override Commander.commands() to add.', name)
      sys.exit(1)

  def _setup_parser(self):
    """
      Sets up the main parser.

      To show version of your customized wst package when --version is invoked, set cls.package_name to your package name.
    """
    self.parser = argparse.ArgumentParser(description=textwrap.dedent(self.__doc__), formatter_class=argparse.RawDescriptionHelpFormatter)
    self.parser.register('action', 'parsers', AliasedSubParsersAction)

    versions = []
    for pkg in filter(None, [getattr(self, 'package_name', None), 'workspace-tools']):
      try:
        versions.append('%s %s' % (pkg, pkg_resources.get_distribution(pkg).version))
      except Exception:
        pass

    self.parser.add_argument('-v', '--version', action='version', version='\n'.join(versions))
    self.parser.add_argument('--debug', action='store_true', help='Turn on debug mode')

  def setup_parsers(self):
    """
      Sets up parsers for all commands
    """

    self._setup_parser()

    self.subparsers = self.parser.add_subparsers(title='sub-commands', help='List of sub-commands')
    self.subparsers.remove_parser = lambda *args, **kwargs: _remove_parser(self.subparsers, *args, **kwargs)

    for name, command in sorted(self.commands().items()):
      doc, _ = command.docs()
      help = filter(None, doc.split('\n'))[0]
      aliases = [command.alias] if command.alias else None

      parser = self.subparsers.add_parser(name, aliases=aliases, description=textwrap.dedent(doc), help=help,
                                          formatter_class=argparse.RawDescriptionHelpFormatter)
      parser.set_defaults(command=name)

      cmd_args = command.arguments()

      if isinstance(cmd_args, tuple):
        normal_args, chain_args = cmd_args
      else:
        normal_args = cmd_args
        chain_args = []

      for args, kwargs in normal_args:
        parser.add_argument(*args, **kwargs)

      if chain_args:
        group = parser.add_argument_group('chaining options')
        for args, kwargs in chain_args:
          group.add_argument(*args, **kwargs)


# Copied from https://gist.github.com/sampsyo/471779
class AliasedSubParsersAction(argparse._SubParsersAction):

  class _AliasedPseudoAction(argparse.Action):
    def __init__(self, name, aliases, help):
      dest = name
      if aliases:
        dest += ' (%s)' % ','.join(aliases)
      sup = super(AliasedSubParsersAction._AliasedPseudoAction, self)
      sup.__init__(option_strings=[], dest=dest, help=help)

  def add_parser(self, name, **kwargs):
    if 'aliases' in kwargs:
      aliases = kwargs['aliases'] or []
      del kwargs['aliases']
    else:
      aliases = []

    parser = super(AliasedSubParsersAction, self).add_parser(name, **kwargs)

    # Make the aliases work.
    for alias in aliases:
      self._name_parser_map[alias] = parser
    # Make the help text reflect them, first removing old help entry.
    if 'help' in kwargs:
      help = kwargs.pop('help')
      self._choices_actions.pop()
      pseudo_action = self._AliasedPseudoAction(name, aliases, help)
      self._choices_actions.append(pseudo_action)

    return parser


def _remove_parser(self, name, **kwargs):
  # remove choice help from list
  if 'help' in kwargs:
    alias_name = '%s (' % name
    self._choices_actions = [action
                             for action in self._choices_actions
                             if action.dest != name and not action.dest.startswith(alias_name)]

  # remove the parser from the map
  self._name_parser_map.pop(name, None)

  # remove aliases
  aliases = kwargs.pop('aliases', ())
  for alias in aliases:
    self._name_parser_map.pop(alias, None)



class AbstractCommand(object):
  """
    Change the docstring of the command class to set the description of the command.
  """
  #: CLI command alias
  alias = None

  def __init__(self, **kwargs):
    self.__dict__ = kwargs

  def __getattr__(self, attr):
    try:
      super(AbstractCommand, self).__getattr__(attr)
    except Exception:
      return None

  @classmethod
  def name(cls):
    """ Name of the command """
    return cls.__name__.lower()

  @classmethod
  def make_args(cls, *args, **kwargs):
    """ Helper method to compose arguments for :meth:`arguments` """
    return args, kwargs

  @classmethod
  def docs(cls):
    """
    Split the param description from from the class's __doc__

    :return: Tuple of (help_doc, param_docs) where help_doc is the part before :param, and param_docs is a dict of param -> param doc.
    :rtype: tuple(str, dict(str, str))
    """
    doc_parts = cls.__doc__ and cls.__doc__.split(':param ')

    # Try to get / merge from parent class
    if not doc_parts or super(cls, cls).__doc__ != cls.__doc__:
      doc_parts_old = super(cls, cls).__doc__.split(':param ')
      # Merge parent class docs (just the params part)
      if not doc_parts:
        doc_parts = doc_parts_old
      else:
        doc_parts.extend(doc_parts_old[1:])  # params
        if not doc_parts[0].strip():      # description
          doc_parts[0] = doc_parts_old[0]

    doc = doc_parts[0].rstrip()
    params = {}

    for doc_part in doc_parts[1:]:
      type_param, param_doc = doc_part.split(':', 1)
      param = type_param.split()[-1]
      params[param] = param_doc.strip()

    return doc, params

  @classmethod
  def arguments(cls):
    """
      List of arguments to be passed to argparse's add_argument.
      Optionally, return a tuple of two lists where first one consists
      of normal arguments, and the 2nd one consists of chaining arguments.

      Use :meth:`make_args` to create the args::

        return [cls.make_args('--name', help='Product name')]

      Or return a tuple of two lists where the 2nd one consists of chaining options::

        return ([cls.make_args('--name', help='Product name')],
                [cls.make_args('--test', help='Run tests after commit')])


      :return: List of add_argument params or tuple of two lists
    """
    return []

  def run(self):
    raise NotImplementedError('%s has not implemented run()' % self.__class__.__name__)



from contextlib import contextmanager
import os
import shutil
from tempfile import mkdtemp

from workspace.utils import run


@contextmanager
def temp_dir():
  try:
    cwd = os.getcwd()
    dtemp = mkdtemp()
    os.chdir(dtemp)

    yield dtemp

  finally:
    os.chdir(cwd)
    shutil.rmtree(dtemp)


@contextmanager
def temp_git_repo():
  with temp_dir() as dir:
    run('git init')
    yield dir


@contextmanager
def temp_remote_git_repo():
  with temp_dir() as dir:
    run('git clone https://github.com/maxzheng/remoteconfig.git')
    repo_dir = os.path.join(dir, 'remoteconfig')
    os.chdir(repo_dir)
    yield repo_dir



from glob import glob
import logging
import os
from operator import itemgetter
import sys

from tabulate import tabulate

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.scm import product_name, is_git_repo, current_branch, product_path, checkout_branch, repo_path
from workspace.utils import background_processes, run, run_in_background

log = logging.getLogger(__name__)


class Wait(AbstractCommand):
  """
    Wait for an event to be completed and optionally start background/waiting tasks.

    If no argument is passed, then show running background/waiting tasks.

    Extra arguments passed to wait will be run in the background. I.e. ws wait sleep 10.
    This can be used with any of the waiting or chaining options.  If any chaining options
    are used, then extra argument commands will run afterward/last.

    :param int review: Wait for 'Ship It' from review board.
                       This is blocking, and so can be used to chain commands in command prompt.
                       i.e.: ws wait -r && ws push (equivalent of ws wait -p)
    :param bool publish: Wait for a new version to be published.
                         This is blocking, and so can be used to chain commands in command prompt.
    :param int interval: Minutes to wait between each check.
                         Defaults to 5 minutes.
    :param bool push: Wait for review and push. Implies --review, however it is non-blocking / runs in background.
    :param list bump_in: Wait for publish and bump it in the given products/product groups.
                    Implies --publish, however it is non-blocking / runs in the background.
    :param bool log: List all wait logs in temp dir (/var/tmp/workspace/wait-*), optionally matching current repo if
                     in a repo.
                     Use -ll to view last log. Each additional 'l' will view the log before last.
    :param list extra_args: Arbitrary commands to run in background.
  """
  alias = 'w8'

  def __init__(self, *args, **kwargs):
    #: Run wait in background if possible
    self.in_background = False

    super(Wait, self).__init__(*args, **kwargs)

    if self.push:
      self.review = True

    if self.bump_in:
      self.publish = True

    if not self.in_background:
      self.in_background = self.push or self.bump_in or self.extra_args

    #: Save the branch as it could change while waiting
    self.branch = is_git_repo() and current_branch()

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return ([
        cls.make_args('-r', '--review', action='store_true', help=docs['review']),
        cls.make_args('-P', '--publish', action='store_true', help=docs['publish']),
        cls.make_args('-i', '--interval', type=int, default=5, help=docs['interval']),
        cls.make_args('-l', '--log', action='count', help=docs['log'])
      ],
      [
        cls.make_args('-p', '--push', action='store_true', help=docs['push']),
        cls.make_args('-bi', '--bump-in', metavar='PRODUCT', nargs='*', help=docs['bump_in']),
      ])

  def run(self):
    # Refactor to have core logic here and call out to self._wait_for_review and self._wait_for_publish
    if type(self) == Wait and (self.review or self.publish):
      raise NotImplementedError('Not implemented. Please implement Wait.run() in a subclass with review/publish')

    if not (self.review or self.publish or self.log or self.extra_args):
      processes = background_processes()
      if processes:
        print tabulate(sorted(processes, key=itemgetter(0, 1)), headers=['Repo', 'Task', 'PID'])
      sys.exit(0)

    if self.log:
      repo = repo_path()
      logs = sorted(glob('/var/tmp/workspace/wait-*'), key=lambda f: os.stat(f).st_mtime)

      if len(logs) > 10:
        for log_file in logs[:len(logs)-10]:
          log.debug('Deleting old log: %s', log_file)
          os.unlink(log_file)
        logs = logs[-10:]

      if repo:
        name = product_name(repo)
        repo_key = 'wait-' + name.replace('-', '_') + '-'
        logs = [l for l in logs if repo_key in l]

      if logs:
        if self.log == 1:
          print '\n'.join(logs)
        else:
          print open(logs[-(self.log-1)]).read()

      else:
        print 'No logs found'

    if self.push:
      self.commander.run('push', branch=self.branch)

    if self.bump_in:
      name = product_name()
      for product in expand_product_groups(self.bump_in):
        path = product_path(product)
        if os.path.exists(path):
          if not os.fork():  # child
            os.chdir(path)
            checkout_branch('master')
            self.commander.run('bump', rb=True, test=True, push=True, names=[name])

    if self.extra_args:
      run_in_background(' '.join(self.extra_args))
      run(self.extra_args, shell=True)

  @property
  def then_action(self):
    """ Description for the next action if any. """
    action = None

    if self.push:
      action = 'push'

    if self.bump_in:
      action = 'bump in %s' % ', '.join(self.bump_in)

    if self.extra_args:
      action = ' '.join(self.extra_args)

    return ', then ' + action if action else ''



import logging
import os
import re
import shutil
import sys

import requests

from workspace.config import config
from workspace.utils import run, silent_run, parent_path_with_dir


log = logging.getLogger(__name__)

USER_REPO_REFERENCE_RE = re.compile('^[\w-]+/[\w-]+$')
SVN_COMMIT_HEADER_RE = re.compile('^(-+|r\d+\s.*\slines?|\s+)$')
GITIGNORE_FILE = '.gitignore'
GITIGNORE = """\
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]

# C extensions
*.so

# Distribution / packaging
.Python
env/
build/
develop-eggs/
dist/
downloads/
eggs/
lib/
lib64/
parts/
sdist/
var/
*.egg-info/
.installed.cfg
*.egg

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.coverage
.cache
nosetests.xml
coverage.xml

# Translations
*.mo
*.pot

# Django stuff:
*.log

# Sphinx documentation
docs/_build/

# PyBuilder
target/
"""


def workspace_path():
  """ Guess the workspace path based on if we are in a repo or not. """
  repo_path = is_repo()
  if repo_path:
    return os.path.dirname(repo_path)
  else:
    return os.getcwd()


def commit_logs(limit=None, repo=None, diff=False, show_revision=None, extra_args=None, to_pager=False):
  if show_revision:
    diff = True
    if not limit:
      limit = 1

  if is_git_repo(repo):
    cmd = ['git', 'log', '--decorate']
    if show_revision:
      cmd.extend(['-U', show_revision])
    if limit:
      cmd.append('-%d' % limit)
    if diff:
      cmd.append('-p')
    if extra_args:
      cmd.extend(extra_args)

  else:
    cmd = ['svn', 'log']
    if show_revision:
      cmd.extend(['-r', show_revision])
    if limit:
      cmd.extend(['-l', str(limit)])
    if diff:
      cmd.append('--diff')
    if extra_args:
      limit_re = re.compile('-\d')
      for i, arg in enumerate(extra_args):
        if limit_re.match(arg):
          extra_args[i] = '-l %s' % arg.lstrip('-')
      cmd.extend(extra_args)
    if to_pager and (show_revision or not limit or limit > 3):
      cmd.extend(['|', os.environ.get('PAGER') or 'less'])

  return run(cmd, return_output=not to_pager, shell=to_pager, cwd=repo)


def add_files(repo=None, files=None):
  if not repo:
    repo = git_repo_path()

  if files:
    files = ' '.join(files)
  else:
    files = '.'

  silent_run('git add --all ' + files, cwd=repo)


def repo_check():
  if not is_repo():
    log.error('This should be run from within a product checkout')
    sys.exit(1)


def git_repo_check(hint='This command only supports git repo.'):
  if not is_git_repo():
    log.error('Not a git repository. %s', hint)
    if is_svn_repo():
      log.info('Please consider using git-svn for your svn repo as git is AWESOME!! :)')
    sys.exit(1)


def extract_commit_msgs(output, is_git=True):
  """ Returns a list of commit msgs from the given output. """
  msgs = []

  if output:
    msg = []

    for line in output.split('\n'):
      if not line:
        continue

      if is_git:
        is_commit_msg = line.startswith(' ') or not line
      else:
        is_commit_msg = not(SVN_COMMIT_HEADER_RE.match(line))

      if is_commit_msg:
        if is_git and line.startswith('    '):
          line = line[4:]
        msg.append(line)
      elif msg:
        msgs.append('\n'.join(msg))
        msg = []

    if msg:
      msgs.append('\n'.join(msg))

  return msgs


def is_git_svn_repo(path=None):
  """ Checks if given or current path is a git svn repo """
  repo_path = git_repo_path(path)
  return repo_path and os.path.isfile(os.path.join(repo_path, '.git', 'svn', '.metadata'))


def is_repo(path=None):
  return repo_path(path)


def repo_path(path=None):
  # Last element is longest path, which is the current repo in a nested repo situation.
  return sorted([git_repo_path(path), svn_repo_path(path)])[-1]


def is_svn_repo(path=None):
  svn_repo = svn_repo_path(path)
  return svn_repo and svn_repo == repo_path(path)


def svn_repo_path(path=None):
  """ Checks if given or current path is a svn repo """
  return parent_path_with_dir('.svn', path)


def is_git_repo(path=None):
  git_repo = git_repo_path(path)
  return git_repo and git_repo == repo_path(path)


def git_repo_path(path=None):
  """ Checks if given or current path is a git repo """
  return parent_path_with_dir('.git', path)


def product_repos():
  """ Product repos for the current workspace. """
  return repos(workspace_path())


def repos(dir=None):
  """ Returns a list of repos either for the given directory or current directory or in sub-directories. """
  repos = []
  cwd = dir or os.getcwd()

  if is_repo(cwd):
    repos.append(repo_path(cwd))
    return repos

  for dir in os.listdir(cwd):
    path = os.path.join(cwd, dir)
    if os.path.isdir(path) and is_repo(path):
      repos.append(path)

  return repos


def checkout_branch(branch, repo_path=None):
  """ Checks out the branch in the given or current repo. Raises on error. """
  silent_run(['git', 'checkout', branch], cwd=repo_path)


def create_branch(branch, from_branch=None):
  """ Creates a branch from the current branch. Raises on error """
  cmd = ['git', 'checkout', '-b', branch]
  if from_branch:
    cmd.append(from_branch)
  silent_run(cmd)


def update_branch(repo=None):
  silent_run('git rebase master', cwd=repo)


def remove_branch(branch, raises=False):
  """ Removes branch """
  silent_run(['git', 'branch', '-D', branch], raises=raises)


def rename_branch(branch, new_branch):
  silent_run(['git', 'branch', '-m', branch, new_branch])


def merge_branch(branch, squash=False):
  cmd = ['git', 'merge', branch]
  if squash:
    cmd.append('--squash')
  silent_run(cmd)


def diff_branch(right_branch, left_branch='master', path=None):
  cmd = 'git log %s..%s' % (left_branch, right_branch)

  return run(cmd, cwd=path, return_output=True)


def all_branches(repo=None):
  """ Returns all branches. The first element is the current branch. """
  branch_output = silent_run('git branch', cwd=repo, return_output=True)
  branches = []

  if branch_output:
    for branch in branch_output.split('\n'):
      branch = branch.strip()
      if branch.startswith('*'):
        branches.insert(0, branch.strip('*').strip())
      elif branch:
        branches.append(branch)

  return branches


def current_branch(repo=None):
  return all_branches(repo)[0]


def update_repo(path=None):
  """ Updates given or current repo to HEAD """

  if is_git_svn_repo(path):
    silent_run('git svn rebase', cwd=path)
  elif is_git_repo(path):
    silent_run('git pull --rebase', cwd=path)
  else:
    silent_run('svn update --ignore-externals', cwd=path)


def push_repo(path=None):
  if is_git_svn_repo(path):
    silent_run('git svn dcommit')
  else:
    silent_run('git push')


def stat_repo(path=None, return_output=False):
  if is_git_repo(path):
    cmd = 'git status'
  else:
    cmd = 'svn status'

  return run(cmd, cwd=path, return_output=return_output)


def diff_repo(path=None, branch=None, file=None, return_output=False, name_only=False):
  if is_git_repo(path):
    cmd = ['git', 'diff']
    if branch:
      cmd.append(branch)
    if file:
      cmd.append(file)
    if name_only:
      cmd.append('--name-only')

  else:
    cmd = ['svn', 'diff']
    if file:
      cmd.append(file)

  return run(cmd, cwd=path, return_output=return_output)


def commit_changes(msg):
  """ Commits any modified or new files with given message. Raises on error """
  silent_run(['git', 'commit', '-am', msg])
  log.info('Committed change.')


def local_commit(msg=None, amend=False, empty=False):
  cmd = ['git', 'commit']
  if amend:
    cmd.append('--amend')
  if empty:
    cmd.append('--allow-empty')
  if msg:
    cmd.extend(['-m', msg])
  run(cmd)


def svn_revision_range(repo=None, num_commmits=1):
  cmd = ['svn', 'log', '-l', str(num_commmits)]

  if repo and (repo.startswith('http://') or repo.startswith('https://') or repo.startswith('svn+ssh://')):
    cmd.append(repo)
    repo = None

  output = silent_run(cmd, cwd=repo, return_output=True, raises=False)

  if output:
    try:
      output = output.split('\n')
      svn_revision_re = re.compile('r(\d+) ')
      head_revision = svn_revision_re.match(output[1]).group(1)
      for i in range(len(output) - 1, -1, -1):
        match = svn_revision_re.match(output[i])
        if match:
          from_revision = match.group(1)
          break
      return int(from_revision), int(head_revision)
    except Exception:
      return None, None


def clone_svn_repo(product_url, checkout_path, clone_svn_commits):
  prod_name = product_name(product_url)

  log.info('Cloning last %d commit(s) for svn repo using git-svn.', clone_svn_commits)
  if clone_svn_commits > 1:
    log.info('This might take some time...')

  from_revision, head_revision = svn_revision_range(product_url, clone_svn_commits)

  if not head_revision:
    raise Exception('Unable to get latest revision from %s' % product_url)

  workspace_path = os.path.dirname(checkout_path)
  revision_range = '-r%d:%s' % (from_revision, head_revision)

  clone_cmd = ['git', 'svn', 'clone', revision_range]
  if product_url.endswith('/trunk'):
    clone_cmd.extend(['-T', 'trunk', product_url[:-6], checkout_path])
  else:
    clone_cmd.extend([product_url, checkout_path])

  silent_run(clone_cmd, cwd=workspace_path)

  gitignore_file = os.path.join(workspace_path, prod_name, GITIGNORE_FILE)
  if not os.path.exists(gitignore_file):
    with open(gitignore_file, 'w') as fp:
      fp.write(GITIGNORE)
    log.info('Created %s/%s. Please check that in so git ignores build/temp files.', prod_name, os.path.basename(gitignore_file))


def checkout_product(product_url, checkout_path):
  """ Checks out the product from url. Raises on error """
  product_url = product_url.strip('/')

  clone_svn_commits = config.checkout.use_gitsvn_to_clone_svn_commits
  prod_name = product_name(product_url)

  if os.path.exists(checkout_path):
    log.debug('%s is already checked out. Updating...', prod_name)
    if is_git_repo(checkout_path):
      checkout_branch('master', checkout_path)
    return update_repo(checkout_path)

  if re.match('[\w-]+$', product_url):
    try:
      logging.getLogger('requests').setLevel(logging.WARN)
      response = requests.get(config.checkout.search_api_url, params={'q': product_url}, timeout=10)
      response.raise_for_status()
      results = response.json()['items']
      if not results:
        log.error('No repo matching "%s" found.', product_url)
        sys.exit(1)
      product_url = results[0]['ssh_url']
      log.info('Using repo url %s', product_url)
    except Exception as e:
      log.error('Could not find repo for %s using %s due to error: ', product_url, config.checkout.search_api_url, e)
      sys.exit(1)

  elif USER_REPO_REFERENCE_RE.match(product_url):
    product_url = config.checkout.user_repo_url % product_url

  if clone_svn_commits and not product_url.endswith('.git'):
    try:
      return clone_svn_repo(product_url, checkout_path, clone_svn_commits)
    except Exception as e:
      log.exception('Falling back to svn as git-svn clone failed: %s', e)
      if os.path.exists(checkout_path):
        shutil.rmtree(checkout_path)

  if product_url.endswith('.git'):
    silent_run(['git', 'clone', product_url, checkout_path])
  else:
    silent_run(['svn', 'checkout', product_url, checkout_path])


def checkout_files(files, repo_path=None):
  """ Checks out the given list of files. Raises on error. """
  if is_git_repo(repo_path):
    silent_run(['git', 'checkout'] + files, cwd=repo_path)
  else:
    silent_run(['svn', 'revert'] + files, cwd=repo_path)


def hard_reset(to_commit):
  run(['git', 'reset', '--hard', to_commit])


def product_name(product_url=None):
  if not product_url:
    product_url = repo_path()

  product_url = product_url.strip('/')

  if product_url.endswith('.git'):
    name = product_url[:-4]
  elif product_url.endswith('_trunk') or product_url.endswith('/trunk'):
    name = product_url[:-6]
  else:
    name = product_url

  return os.path.basename(name)


def product_checkout_path(product_url, workspace_dir=None):
  return product_path(product_name(product_url), workspace_dir)


def product_path(name, workspace_dir=None):
  if not workspace_dir:
    workspace_dir = workspace_path()

  return os.path.join(workspace_dir, name)


def repo_url(path=None, name='origin', action='push'):
  """
    :param str path: Local repo path
    :param str source: Remote name
    :param str action: Action for the corresponding URL
    :return: Remote url for repo or None if not found
  """
  cmd = 'git remote -v'

  output = run(cmd, cwd=path, return_output=True)

  for line in output.split('\n'):
    if line:
      remote_name, remote_url, remote_action = line.split()
      if remote_name == name and action in remote_action:
        return remote_url



import pytest
import subprocess

from workspace.utils import run


@pytest.mark.parametrize("script", ['wst'])
def test_script_sanity(script):

  try:
    run([script, '-h'], stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError as e:
    print e.output
    assert e.returncode == 0



import logging
import signal

from workspace.commands import AbstractCommand
from workspace.scm import commit_logs, repo_check

log = logging.getLogger(__name__)


class Log(AbstractCommand):
  """
    Show commit logs.

    Extra arguments are passed to the underlying SCM's log command.

    :param bool diff: Generate patch / show diff
    :param str show: Show specific revision. This implies --diff and limit of 1
    :param int limit: Limit number of log entries
    :param list extra_args: Extra args to pass to the underlying SCM's log command
  """

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [
      cls.make_args('-p', '--diff', action='store_true', help=docs['diff']),
      cls.make_args('-r', '--show', help=docs['show']),
      cls.make_args('-n', '--limit', metavar='NUM', type=int, help=docs['limit'])
    ]

  def run(self):

    repo_check()

    # Interrupt for git log results in bad tty
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
      commit_logs(self.limit, diff=self.diff, show_revision=self.show, extra_args=self.extra_args, to_pager=True)
    except Exception as e:
      # Oddly, git log returns non-zero exit whenever user exits while it is still printing
      if self.debug:
        log.exception(e)



from contextlib import contextmanager
import logging
import os
import re
import signal
import subprocess
import sys
import tempfile

import psutil
from setproctitle import setproctitle


log = logging.getLogger(__name__)


def prompt_with_editor(instruction):
  """ Prompt user with instruction in $EDITOR and return the response """

  with tempfile.NamedTemporaryFile(suffix='.tmp') as fh:
    fh.write('\n\n# ' + '\n# '.join(instruction.split('\n')))
    fh.flush()
    editor = os.environ.get('EDITOR', 'vim')
    run([editor, fh.name])
    return '\n'.join(filter(lambda l: not l.startswith('#'), open(fh.name).read().split('\n'))).strip()


def parent_path_with_dir(directory, path=None):
  """
  Find parent that contains the given directory.

  :param str directory: Directory to look for
  :param str path: Initial path to look from. Defaults to current working directory.
  :return: Parent path that contains the directory
  :rtype: str on success or False on failure
  """
  if path == '/':
    return False

  if not path:
    path = os.getcwd()

  if os.path.isdir(os.path.join(path, directory)):
    return path

  return parent_path_with_dir(directory, os.path.dirname(path))


@contextmanager
def log_exception(title=None, exit=False, call=None, stack=False):
  """
  Context generator that logs exceptions as error

  :param str title: Title to log before the exception
  :param bool exit: Do sys.exit when exception occurs
  :param callable call: Call to make before exit
  :param bool stack: Log stacktrace
  """
  try:
    yield
  except (Exception, KeyboardInterrupt) as e:
    if title:
      log.error(title)
    if stack:
      log.exception(e)
    elif str(e):
      log.error(e)
    if call:
      call()
    if exit:
      sys.exit(1)


class RunError(Exception):
  pass


def silent_run(*args, **kwargs):
  """ Same as run with slient=True """
  return run(*args, silent=True, **kwargs)


def run(cmd, cwd=None, silent=None, return_output=False, raises=True, **subprocess_args):
  """
  Runs a CLI command.

  :param list/str cmd: Command with args to run.
  :param str cwd: Change directory to cwd before running
  :param bool/int silent: Suppress stdout/stderr.
                          If True, completely silent.
                          If 2, print cmd output on error.
  :param bool return_output: Return the command output. Defaults silent=True. Set silent=False to see output.
                             If True, always return output.
                             If set to 2, return a tuple of (output, success) where output is the output of the command
                             and success is exit code 0.
                             When used, it is guaranteed to always return output / other options are ignored (like raises).
  :param bool raises: Raise an exception if command exits with an error code.
  :param dict subprocess_args: Additional args to pass to subprocess
  :return: Output or boolean of success depending on option selected
  :raise RunError: if the command exits with an error code and raises=True
  """

  if isinstance(cmd, basestring):
    cmd = cmd.split()

  cmd_str = ' '.join(cmd)

  if 'shell' in subprocess_args and subprocess_args['shell']:
    cmd = cmd_str

  log.debug('Running: %s %s', cmd_str, '[%s]' % cwd if cwd else '')

  if return_output and silent is None:
      silent = True

  try:
    if silent or return_output:
      p = subprocess.Popen(cmd, cwd=cwd, bufsize=0, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **subprocess_args)
      exit_code = -1

      if silent:
        output, _ = p.communicate()
        exit_code = p.returncode
      else:
        output = ''
        ch = True
        while ch:
          ch = p.stdout.read(1)
          sys.stdout.write(ch)
          sys.stdout.flush()
          output += ch
          if p.poll() is not None and exit_code == -1:
            exit_code = p.returncode

      if return_output is True:
        return output
      elif return_output == 2:
        return output, exit_code == 0

      if exit_code == 0:
        if return_output:
          return output
        else:
          return True

      elif raises or silent == 2:
        if output and silent:
          print output.strip()

    else:
      exit_code = subprocess.call(cmd, cwd=cwd, **subprocess_args)
      if exit_code == 0:
        return True

  except Exception as e:
    if raises:
      log.debug(e, exc_info=True)
      raise RunError('Command "%s" could not be run because %s' % (cmd_str, e))

  # We only get here if exit code != 0
  if raises:
    raise RunError('Command "%s" returned non-zero exit status %d' % (cmd_str, exit_code))

  return False


def parallel_call(call, args, callback=None, workers=10, show_progress=None, progress_title='Progress'):
  """
  Call a callable in parallel for each arg

  :param callable call: Callable to call
  :param list(iterable|non-iterable) args: List of args to call. One call per args.
  :param callable callback: Callable to call for each result.
  :param int workers: Number of workers to use.
  :param bool/str/callable: Show progress.
                            If callable, it should accept two lists: completed args and all args and return progress string.
  :return dict: Map of args to their results on completion
  """
  from multiprocessing import Pool, TimeoutError

  signal.signal(signal.SIGTERM, lambda *args: sys.exit(1))
  pool = Pool(workers, lambda: signal.signal(signal.SIGINT, signal.SIG_IGN))

  def to_tuple(a):
    return a if isinstance(a, (list, tuple, set)) else [a]

  try:
    async_results = [(arg, pool.apply_async(call, to_tuple(arg), callback=callback)) for arg in args]

    results = {}
    while len(results) != len(async_results):
      for arg, result in async_results:
        if arg not in results:
          try:
            # This allows processes to be interrupted by CTRL+C
            results[arg] = result.get(1)
          except TimeoutError:
            pass
          except Exception as e:
            results[arg] = str(e)

      if show_progress:
        if callable(show_progress):
          progress = show_progress(results.keys(), args)
        else:
          progress = '%.2f%% completed' % (len(results) * 100.0 / len(async_results))
        show_status('%s: %s' % (progress_title, progress))

    pool.close()
    pool.join()

    return results

  except KeyboardInterrupt:
    os.killpg(os.getpid(), signal.SIGTERM)  # Kills any child processes from subprocesses.
    pool.terminate()
    pool.join()
    sys.exit()


def show_status(message):
  """
    :param str message: Status message to show. If not, then status bar will be cleared.
  """
  enabled = (sys.stdout.isatty() and 'TERM' in os.environ) or os.environ.get('PYCHARM_HOSTED')
  if not enabled:
    return

  sys.stdout.write('%s\r' % message)
  sys.stdout.flush()


def background_processes():
  """
    List of background processes from `run_in_background`
  """
  prog_prefix = os.path.basename(sys.argv[0]) + ' ['
  processes = []

  for process in psutil.process_iter():
    try:
      if process.cmdline()[0].startswith(prog_prefix):
        cmdline = ' '.join(process.cmdline())[len(prog_prefix)-1:]
        repo, task = cmdline.lstrip('[').split(']', 1)
        processes.append((repo, task, process.pid))
    except Exception:
      pass

  return processes


def run_in_background(title, repo=None, info_suffix='[To check, run: {prog} wait]', log_file=None):
  """
    Run any code after this point in the background. This call is idempotent as subsequent calls
    won't do anything except change the title.

    :param str title: Title to set the running process. This should be informative to the user on
                      what is being run in the background and will happen.
    :param str repo: Name of repo the task is being acted on. Defaults to os.getcwd()
    :param str info_suffix: Informational suffix to append when showing the title before forking.
                            {prog} will be replaced by the running program's name.
    :param str log_file: Full path to log file to save output/error. Saves to temp dir by default.
                         Set to "/dev/null" to discard output.
  """
  prog = os.path.basename(sys.argv[0])
  setproctitle('%s [%s] %s' % (prog, os.path.basename(repo or os.getcwd()), title))

  if hasattr(run_in_background, 'forked'):
    return
  else:
    run_in_background.forked = True

  if not log_file:
    def rs(s):  # Remove special characters
      return re.sub('[^a-zA-Z0-9]', '_', s)
    log_file = os.path.join(tempfile.gettempdir(), 'wait-{0}-{1}.out'.format(rs(os.path.basename(repo or os.getcwd())), rs(title)))

  log.info('%s %s' % (title, info_suffix.format(prog=prog)))
  log.debug('Log file: %s', log_file)

  try:
    pid = os.fork()
    if pid > 0:
      sys.exit(0)
  except OSError as e:
    log.error(e)
    sys.exit(1)

  if repo:
    os.chdir(repo)
  os.setsid()
  os.umask(0)

  try:
    pid = os.fork()
    if pid > 0:
      sys.exit(0)
  except OSError as e:
    log.error(e)
    sys.exit(1)

  sys.stdout.flush()
  sys.stderr.flush()
  so = file(log_file, 'w+')
  se = file(log_file, 'w+', 0)
  si = file(log_file, 'r')
  os.dup2(si.fileno(), sys.stdin.fileno())
  os.dup2(so.fileno(), sys.stdout.fileno())
  os.dup2(se.fileno(), sys.stderr.fileno())

  print '[%s] %s' % (os.path.basename(repo or os.getcwd()), title)
  sys.stdout.flush()



import logging
import os

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.scm import checkout_product, checkout_branch, is_git_repo, all_branches, checkout_files, is_repo, product_checkout_path, product_name
log = logging.getLogger(__name__)


class Checkout(AbstractCommand):
  """
    Checkout products (repo urls) or branch, or revert files.

    :param list target: List of products (git/svn repository URLs) to checkout. When inside a git repo,
                        checkout the branch or revert changes for file(s). When inside a svn repo, revert
                        changes for file(s)
  """
  alias = 'co'

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [cls.make_args('target', nargs='+', help=docs['target'])]

  def run(self):
    if is_repo():
      if is_git_repo() and len(self.target) == 1 and self.target[0] in all_branches():
        checkout_branch(self.target[0])
        log.info('Switched to branch %s', self.target[0])
      else:
        checkout_files(self.target)
      return

    product_urls = expand_product_groups(self.target)

    for product_url in product_urls:
      product_url = product_url.strip('/')

      product_path = product_checkout_path(product_url)

      if os.path.exists(product_path):
        log.info('Updating %s', product_name(product_path))
      else:
        log.info('Checking out %s', product_url)

      checkout_product(product_url, product_path)



import logging
import os

from workspace.commands import AbstractCommand
from workspace.commands.helpers import ProductPager
from workspace.scm import diff_repo, repos, product_name, current_branch, is_git_repo
from workspace.utils import log_exception

log = logging.getLogger(__name__)


class Diff(AbstractCommand):
  """
    Show diff on current product or all products in workspace

    :param str file: Show diff for file only
    :param bool master: Diff against the master branch
    :param bool name_only: List file names only. Git only.
  """
  alias = 'di'

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [
      cls.make_args('file', nargs='?', help=docs['file']),
      cls.make_args('-m', '--master', action='store_true', help=docs['master']),
      cls.make_args('-l', '--name-only', action='store_true', help=docs['name_only'])
    ]

  def run(self):
    if self.file:
      scm_repos = [os.getcwd()]
    else:
      scm_repos = repos()

    optional = len(scm_repos) == 1
    pager = ProductPager(optional=optional)

    for repo in scm_repos:
      with log_exception():
        branch = 'master' if self.master else None
        output = diff_repo(repo, branch=branch, file=self.file, return_output=True, name_only=self.name_only)
        if output:
          branch = current_branch(repo) if is_git_repo(repo) else None
          pager.write(product_name(repo), output, branch)

    pager.close_and_wait()



import os
import logging

from workspace.commands import AbstractCommand
from workspace.commands.helpers import ProductPager
from workspace.scm import stat_repo, repos, product_name, all_branches, is_git_repo, is_repo

log = logging.getLogger(__name__)


class Status(AbstractCommand):
  """ Show status on current product or all products in workspace """
  alias = 'st'

  def run(self):

    try:
      scm_repos = repos()
      in_repo = is_repo(os.getcwd())
      optional = len(scm_repos) == 1
      pager = ProductPager(optional=optional)

      for repo in scm_repos:
        stat_path = os.getcwd() if in_repo else repo
        output = stat_repo(stat_path, True)
        nothing_to_commit = 'nothing to commit' in output and 'Your branch is ahead of' not in output

        branches = all_branches(repo) if is_git_repo(repo) else []

        if len(branches) > 1:
          if nothing_to_commit:
            output = '# Branches: %s' % ' '.join(branches)
            nothing_to_commit = False
          elif len(branches) > 1:
            output = '# Branches: %s\n#\n%s' % (' '.join(branches), output)

        if output and not nothing_to_commit:
            pager.write(product_name(repo), output)
    except Exception as e:
      log.error(e)
    finally:
      pager.close_and_wait()



import logging
import re
import sys

from workspace.config import config

from workspace.commands import AbstractCommand
from workspace.scm import local_commit, add_files, git_repo_check, checkout_branch,\
    create_branch, all_branches, diff_branch, current_branch, remove_branch, hard_reset, commit_logs
from workspace.utils import prompt_with_editor

log = logging.getLogger(__name__)


class Commit(AbstractCommand):
  """
    Commit all changes locally, including new files.

    :param str msg: Commit message. The first few words are used to create the branch name if branch isn't provided.
                    That behavior can be configured with [commit] auto_branch_from_commit_words
    :param str branch: Use specified branch for commit instead of auto-computing the branch from commit msg.
    :param bool amend: Amend last commit with any new changes made
    :param bool test: Run tests. Repeat twice (-tt) to test dependents too.
    :param bool push: Push the current branch after commit
    :param bool dummy: Perform a dummy commit without any changes on master branch. This implies --push.
                       Other options are ignored.
    :param int discard: Discard last commit, or branch if there are no more commits. Use multiple times to discard multiple commits.
                        Other options are ignored.
    :param str move: Move last commit to branch. Other options are ignored.
    :param callable test_command: Alternative test command to run
    :param bool skip_auto_branch: Skip automatic branch creation from commit msg
    :param list files: List of files to add instead of all files.
  """
  alias = 'ci'

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return ([
        cls.make_args('msg', nargs='?', help=docs['msg']),
        cls.make_args('-a', '--amend', action='store_true', help=docs['amend']),
        cls.make_args('-d', '--dummy', action='store_true', help=docs['dummy']),
        cls.make_args('-D', '--discard', action='count', help=docs['discard']),
        cls.make_args('--move', metavar='branch', nargs=1, help=docs['move']),
        cls.make_args('-b', '--branch', help=docs['branch'])
      ], [
        cls.make_args('-t', '--test', action='count', help=docs['test']),
        cls.make_args('-r', '--rb', action='store_true',
                      help='Create or update existing RB after commit. Existing RB is looked up in .git/config. '
                           'Use -rt to run / post the test results. Repeat twice to test dependents too (-rtt)'),
        cls.make_args('-p', '--push', action='store_true', help=docs['push'])
      ])

  def run(self):
    git_repo_check()

    if self.dummy:
      checkout_branch('master')
      self.commander.run('update')  # Needs to be updated otherwise empty commit below gets erased in push_branch when update is called
      if not self.msg:
        self.msg = 'Empty commit to trigger build'
      elif not self.msg.startswith('Empty commit'):
        self.msg = 'Empty commit. %s' % self.msg
      local_commit(self.msg, empty=True)
      self.commander.run('push', skip_precommit=True)

    elif self.discard or self.move:
      if not self.branch:
        if self.discard:
          self.branch = current_branch()
        else:
          self.branch = self.move[0]

      if self.discard:
        changes = commit_logs(self.discard) if self.branch == 'master' else diff_branch(self.branch)
      else:
        changes = commit_logs(1)
      changes = filter(None, changes.split('commit '))

      if self.discard and len(changes) <= self.discard and self.branch != 'master':
        checkout_branch('master')
        remove_branch(self.branch, raises=True)
        log.info('Deleted branch %s', self.branch)

      else:
        match = re.match('([a-f0-9]+)(?: \(.*\))\n', changes[0])

        if match:
          last_commit = match.group(1)

          if self.move:
            cur_branch = current_branch()
            create_branch(self.branch)
            checkout_branch(cur_branch)
            log.info('Moved %s to %s', last_commit[:7], self.branch)
            hard_reset(last_commit + '~1')

          else:
            checkout_branch(self.branch)
            hard_reset(last_commit + '~' + str(self.discard))

        else:
          log.error('Odd. No commit hash found in: %s', changes[0])

    else:
      test_output = None

      if not (self.msg or self.amend):
        self.msg = prompt_with_editor('Please provide a commit message. Empty message will cancel the commit.')
        if not self.msg:
          sys.exit()

      if not self.amend and self.test:
        if self.commander.command('test').supports_style_check():
          log.info('Running style check')
          self.commander.run('test', env_or_file=['style'], silent=2)

        log.info('Running tests')
        test_output = self.commander.run('test', return_output=self.rb, test_dependents=self.test > 1)

        if self.rb:
          success, _ = self.commander.command('test').summarize(test_output)

          if not success:
            sys.exit(1)

      branches = all_branches()
      cur_branch = branches and branches[0]

      if (not (not self.rb and self.push or self.amend) and cur_branch == 'master' and not self.branch and self.msg and
         config.commit.auto_branch_from_commit_words):
        self.branch = self._branch_for_msg(self.msg, config.commit.auto_branch_from_commit_words, branches)

      if self.branch:
        if branches:
          if self.branch in branches:
            if self.branch != cur_branch:
              checkout_branch(self.branch)

          else:
            create_branch(self.branch, 'master')

        else:  # Empty repo without a commit has no branches
          create_branch(self.branch)

      add_files(files=self.files)
      local_commit(self.msg, self.amend)

      if self.amend and self.test:
        if self.commander.command('test').supports_style_check():
          log.info('Running style check')
          self.commander.run('test', env_or_file=['style'], silent=2)

        log.info('Running tests')
        test_output = self.commander.run('test', return_output=self.rb, test_dependents=self.test > 1)

        if self.rb:
          success, _ = self.commander.command('test').summarize(test_output)

          if not success:
            sys.exit(1)

      if self.rb:
        publish = self.push and (self.amend or test_output)
        self.commander.run('review', publish=publish, test=test_output, skip_prereview=self.test)

        if self.push:
          if not publish:
            log.info("Review was not published as there is no testing done. Please update testing done and then publish")
          self.branch = current_branch()  # Ensure branch is set for push as it could change while waiting
          self.commander.run('wait', review=True, in_background=True)

      if self.push:
        self.commander.run('push', branch=self.branch, skip_precommit=self.rb)

      return test_output

  @classmethod
  def _branch_for_msg(cls, msg, words=3, branches=None):
    ignored_num_re = re.compile('^\d+$')
    ignored_words = ['and', 'but', 'for', 'from']
    ignored_word_length = 2
    branch_name = []
    word_count = 0

    if msg.startswith('DRAFT: '):
      msg = msg.replace('DRAFT: ', '', 1)

    for word in re.split('[\W\_]+', msg):
      if not word:
        continue

      branch_name.append(word.lower())

      if word not in ignored_words and not ignored_num_re.match(word) and len(word) > ignored_word_length:
        word_count += 1

      if word_count >= words and (not branches or '-'.join(branch_name) not in branches):
        break

    if not branch_name:
      raise Exception('No words found in commit msg to create branch name')

    has_ignored_or_short_name = branch_name[-1] in ignored_words or not ignored_num_re.match(word) and len(branch_name[-1]) <= ignored_word_length
    no_name_conflict = not branches or '-'.join(branch_name[:-1]) not in branches
    if has_ignored_or_short_name and no_name_conflict:
      branch_name = branch_name[:-1]

    branch_name = '-'.join(branch_name)

    if branches and branch_name in branches:
      raise Exception('Branch "%s" already exist.\n\tPlease use a more unique commit message or specify branch with -b. Or '
                      'to amend an existing commit, use -a' % branch_name)

    return branch_name



import argparse
import logging
import os
import pkg_resources
import re
import sys
import tempfile

import simplejson as json

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups, ToxIni
from workspace.config import config
from workspace.scm import repo_check, product_name, repo_path, product_repos, product_path, repos, workspace_path, current_branch
from workspace.utils import run, log_exception, parallel_call

log = logging.getLogger(__name__)

TEST_RE = re.compile('\d+ (?:passed|error|failed|xfailed).* in [\d\.]+ seconds')
BUILD_RE = re.compile('BUILD SUCCESSFUL')


class Test(AbstractCommand):
  """
    Run tests and manage test environments for product.

    Extra optional boolean args (such as -s, -v, -vv, etc) are passed to py.test.

    :param list env_or_file: The tox environment to act upon, or a file to pass to py.test (only used
                             if file exists, we don't need to redevelop, and py.test is used as a command
                             for the default environements). Defaults to the envlist in tox.
    :param str repo: Repo path to test instead of current repo
    :param bool show_dependencies: Show where product dependencies are installed from and their versions.
    :param bool test_dependents: Run tests in this product and in checked out products that depends on this product.
                                 This product must be installed as editable in its dependents for the results to be useful.
                                 Most args are ignored when this is used.
    :param bool redevelop: Redevelop the test environment by installing on top of existing one.
                           This is implied if test environment does not exist, or whenever requirements.txt or
                           pinned.txt is modified after the environment was last updated.
                           Use -ro to do redevelop only without running tests.
                           Use -rr to remove the test environment first before redevelop (recreate).
    :param bool install_only: Modifier for redevelop. Perform install only without running test.
    :param bool match_test: Only run tests with method name that matches pattern
    :param bool return_output: Return test output instead of printing to stdout
    :param str num_processes: Number of processes to use when running tests in parallel
    :param list tox_cmd: Alternative tox command to run.
                         If env is passed in (from env_or_file), '-e env' will be appended as well.
    :param str tox_ini: Path to tox_ini file.
    :param dict tox_commands: Map of env to list of commands to override "[testenv:env] commands" setting for env.
                              Only used when not developing.
    :param list args: Additional args to pass to py.test
    :param bool silent: Run tox/py.test silently. Only errors are printed and followed by exit.
    :param bool debug: Turn on debug logging
    :param list install_editable: List of products or product groups to install in editable mode.
    :param list extra_args: Extra args from argparse to be passed to py.test
    :return: Dict of env to commands ran on success. If return_output is True, return a string output.
             If test_dependents is True, return a mapping of product name to the mentioned results.
  """

  def __init__(self, *args, **kwargs):
    kwargs.setdefault('tox_commands', {})
    kwargs.setdefault('silent', False)  # Enables test output streaming for return_output=True
    super(Test, self).__init__(*args, **kwargs)

  @classmethod
  def supports_style_check(cls, repo=None):
    try:
      if not repo:
        repo = repo_path()
      tox = ToxIni(repo)
      return 'style' in tox.envlist

    except Exception as e:
      log.debug(e)

    return False

  @classmethod
  def summarize(cls, tests, include_no_tests=True):
    """
      Summarize the test results

      :param dict|str tests: Map of product name to test result, or the test result of the current prod.
      :param bool include_no_tests: Include "No tests" results when there are no tests found.
      :return: A tuple of (success, list(summaries)) where success is True if all tests pass and summaries
               is a list of passed/failed summary of each test or just str if 'tests' param is str.
    """
    prod_name = product_name()

    if isinstance(tests, dict):
      product_tests = tests
    else:
      product_tests = {}
      product_tests[prod_name] = tests

    success = True
    summaries = []

    def append_summary(summary, name=None):
      if len(product_tests) == 1:
        summaries.append(summary)
      else:
        summaries.append("%s: %s" % (name, summary))

    for name in sorted(product_tests, key=lambda n: n == prod_name or n):
      if not product_tests[name]:
        success = False
        append_summary('Test failed / No output', name)

      elif product_tests[name] is True:
        append_summary('Test successful / No output', name)

      elif 'collected 0 items' in product_tests[name] and 'error' not in product_tests[name]:
        append_summary('No tests')

      else:
        match = TEST_RE.search(product_tests[name])

        if not match:  # Fall back to build if there are no tests.
          match = BUILD_RE.search(product_tests[name])

        if match:
          append_summary(match.group(0), name)
        else:
          append_summary('No test summary found in output', name)

        summary_lines = [l for l in product_tests[name].replace('xfailed', '').split('\n') if l.startswith('===')]
        if not len(summary_lines) == 2 or 'failed' in summary_lines[-1] or 'error' in summary_lines[-1]:
          success = False

    return success, summaries if isinstance(tests, dict) else summaries[0]

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [
      cls.make_args('env_or_file', nargs='*', help=docs['env_or_file']),
      cls.make_args('-k', metavar='NAME_PATTERN', dest='match_test', help=docs['match_test']),
      cls.make_args('-n', metavar='NUM_PROCESSES', type=int, dest='num_processes', help=docs['num_processes']),
      cls.make_args('-d', '--show-dependencies', metavar='FILTER', action='store', nargs='?', help=docs['show_dependencies'], const=True),
      cls.make_args('-t', '--test-dependents', action='store_true', help=docs['test_dependents']),
      cls.make_args('-r', '--redevelop', action='count', help=docs['redevelop']),
      cls.make_args('-o', action='store_true', dest='install_only', help=argparse.SUPPRESS),
      cls.make_args('-e', '--install-editable', nargs='+', help=docs['install_editable'])
    ]

  def run(self):
    repo_check()

    if self.test_dependents:
      name = product_name()

      # Convert None to list for tuple([])
      if not self.env_or_file:
        self.env_or_file = []
      if not self.extra_args:
        self.extra_args = []

      test_args = (
        ('env_or_file', tuple(self.env_or_file)),
        ('return_output', True),
        ('num_processes', self.num_processes),
        ('silent', True),
        ('debug', self.debug),
        ('extra_args', tuple(self.extra_args))
      )

      test_repos = [repo_path()]
      test_repos.extend(r for r in repos(workspace_path()) if self.product_depends_on(r, name) and r not in test_repos)
      test_args = [(r, test_args, self.__class__) for r in test_repos]

      def test_done(result):
        name, output = result
        success, summary = self.summarize(output)

        if success:
          log.info('%s: %s', name, summary)

        else:
          temp_output_file = os.path.join(tempfile.gettempdir(), 'test-%s.out' % name)
          with open(temp_output_file, 'w') as fp:
            fp.write(output)
          temp_output_file = 'See ' + temp_output_file

          log.error('%s: %s', name, '\n\t'.join([summary, temp_output_file]))

      def show_remaining(completed, all_args):
        completed_repos = set(product_name(r) for r, _, _ in completed)
        all_repos = set(product_name(r) for r, _, _ in all_args)
        remaining_repos = sorted(list(all_repos - completed_repos))
        if len(remaining_repos):
          repo = remaining_repos.pop()
          more = '& %d more' % len(remaining_repos) if remaining_repos else ''
          return '%s %s' % (repo, more)
        else:
          return 'None'

      repo_results = parallel_call(test_repo, test_args, callback=test_done, show_progress=show_remaining, progress_title='Remaining')

      for _, result in repo_results.values():
        success, _ = self.summarize(result)
        if not (success or self.return_output):
            sys.exit(1)

      return dict(repo_results.values())

    if not self.repo:
      self.repo = repo_path()

    # Strip out venv bin path to python to avoid issues with it being removed when running tox
    if 'VIRTUAL_ENV' in os.environ:
      venv_bin = os.environ['VIRTUAL_ENV']
      os.environ['PATH'] = os.pathsep.join([p for p in os.environ['PATH'].split(os.pathsep)
                                            if os.path.exists(p) and not p.startswith(venv_bin)])

    envs = []
    files = []

    if self.env_or_file:
      for ef in self.env_or_file:
        if os.path.exists(ef):
          files.append(os.path.abspath(ef))
        else:
          envs.append(ef)

    pytest_args = ''
    if self.match_test or self.num_processes is not None or files or self.extra_args:
      pytest_args = []
      if self.match_test:
        pytest_args.append('-k ' + self.match_test)
      if self.num_processes is not None:
        pytest_args.append('-n ' + str(self.num_processes))
      if self.extra_args:
        pytest_args.extend(self.extra_args)
      if files:
        pytest_args.extend(files)
      pytest_args = ' '.join(pytest_args)
      os.environ['PYTESTARGS'] = pytest_args

    tox = ToxIni(self.repo, self.tox_ini)

    if not envs:
      envs = tox.envlist

    env_commands = {}

    if self.install_only and not self.redevelop:
      self.redevelop = 1

    if self.show_dependencies:
      for env in envs:
        self.show_installed_dependencies(tox, env, filter_name=self.show_dependencies)

    elif self.install_editable:
      for env in envs:
        if len(envs) > 1:
          print(env + ':')
        self.install_editable_dependencies(tox, env, editable_products=self.install_editable)

    elif self.redevelop:
      if self.tox_cmd:
        cmd = self.tox_cmd
      else:
        cmd = ['tox', '-c', tox.path]

      if envs:
        cmd.extend(['-e', ','.join(envs)])

      if self.redevelop > 1:
        cmd.append('-r')

      if self.install_only:
        cmd.append('--notest')

      output = run(cmd, cwd=self.repo, raises=not self.return_output, silent=self.silent, return_output=self.return_output)

      if not output:
        if self.return_output:
          return False
        else:
          sys.exit(1)

      for env in envs:
        env_commands[env] = ' '.join(cmd)

        # Touch envdir
        envdir = tox.envdir(env)
        if os.path.exists(envdir):
          os.utime(envdir, None)

        # Strip entry version
        self._strip_version_from_entry_scripts(tox, env)

      if self.return_output:
        return output

    else:
      for env in envs:
        envdir = tox.envdir(env)

        def requirements_updated():
          req_mtime = 0
          requirements_files = ['requirements.txt', 'pinned.txt', 'tox.ini']
          for req_file in requirements_files:
            req_path = os.path.join(self.repo, req_file)
            if os.path.exists(req_path):
              req_mtime = max(req_mtime, os.stat(req_path).st_mtime)
          return req_mtime > os.stat(envdir).st_mtime

        if not os.path.exists(envdir) or requirements_updated():
          env_commands.update(
              self.commander.run('test', env_or_file=[env], repo=self.repo, redevelop=True, tox_cmd=self.tox_cmd,
                                 tox_ini=self.tox_ini, tox_commands=self.tox_commands, match_test=self.match_test,
                                 num_processes=self.num_processes, silent=self.silent,
                                 debug=self.debug, extra_args=self.extra_args))
          continue

        if len(envs) > 1 and not self.silent:
            print(env)

        commands = self.tox_commands.get(env) or tox.commands(env)
        env_commands[env] = '\n'.join(commands)

        for command in commands:
          full_command = os.path.join(envdir, 'bin', command)

          command_path = full_command.split()[0]
          if os.path.exists(command_path):
            if 'py.test' in full_command:
              if 'PYTESTARGS' in full_command:
                full_command = full_command.replace('{env:PYTESTARGS:}', pytest_args)
              else:
                full_command += ' ' + pytest_args
            activate = '. ' + os.path.join(envdir, 'bin', 'activate')
            output = run(activate + '; ' + full_command, shell=True, cwd=self.repo, raises=False, silent=self.silent, return_output=self.return_output)
            if not output:
              if self.return_output:
                return False
              else:
                sys.exit(1)

            if env != envs[-1] and not self.silent:
              print()

            if self.return_output:
              return output
          else:
            log.error('%s does not exist', command_path)
            if self.return_output:
              return False
            else:
              sys.exit(1)

    return env_commands

  def _strip_version_from_entry_scripts(self, tox, env):
    """ Strip out version spec "==1.2.3" from entry scripts as they require re-develop when version is changed in develop mode. """
    name = product_name(tox.repo)
    script_bin = tox.bindir(env)

    if os.path.exists(script_bin):
      name_version_re = re.compile('%s==[0-9\.]+' % name)
      removed_from = []
      for script in os.listdir(script_bin):
        script_path = os.path.join(script_bin, script)

        if os.path.isfile(script_path):
          with open(script_path) as fp:
            script = fp.read()

          if name_version_re.search(script):
            new_script = name_version_re.sub(name, script)
            with open(script_path, 'w') as fp:
              fp.write(new_script)
            removed_from.append(os.path.basename(script_path))

      if removed_from:
        log.debug('Removed version spec from entry script(s): %s', ', '.join(removed_from))

  def show_installed_dependencies(self, tox, env, return_output=False, filter_name=None):
    script_template = """
import json
import os
import pip
import sys

# Required params to run this script
package = '%s'
json_output = %s
env = '%s'
filter_name = '%s'

cwd = os.getcwd()
workspace_dir = os.path.dirname(cwd)

try:
  libs = [(p.key, p.version, p.location) for p in pip.get_installed_distributions()]
except Exception as e:
  print(e)
  sys.exit(1)

output = []

if not json_output:
  print(env + ':')

def strip_cwd(dir):
  if dir.startswith(cwd + '/'):
    dir = dir[len(cwd):].lstrip('/')
  elif dir.startswith(workspace_dir):
    dir = os.path.join('..', dir[len(workspace_dir):].lstrip('/'))
  return dir

for lib, version, location in sorted(libs):
  if filter_name and filter_name not in lib:
    continue
  if json_output:
    output.append((lib, version, location))
  else:
    output.append('  %%-25s %%-10s  %%s' %% (lib, version, strip_cwd(location)))

if json_output:
  print(json.dumps(output))
else:
  print('\\n'.join(output))
"""

    name = product_name(tox.repo)
    filter_name = isinstance(filter_name, str) and filter_name or ''
    script = script_template % (name, return_output, env, filter_name)

    python = tox.bindir(env, 'python')

    if not os.path.exists(python):
      log.error('Test environment %s is not installed. Please run without -d / --show-dependencies to install it first.', env)
      sys.exit(1)

    return run([python, '-c', script], return_output=return_output, raises=False)

  def install_editable_dependencies(self, tox, env, editable_products):
    name = product_name(tox.repo)
    editable_products = expand_product_groups(editable_products)

    dependencies_output = self.show_installed_dependencies(tox, env, return_output=True)
    if not dependencies_output:
      log.debug('%s is not installed or there is no dependencies - skipping editable mode changes', name)
      return

    try:
      product_dependencies_list = json.loads(dependencies_output)
    except Exception as e:
      log.debug('Failed to get installed dependencies - skipping editable mode changes: %s', e)
      return

    product_dependencies = {}

    for dep, _, path in product_dependencies_list:
      product_dependencies[dep] = path

    available_products = [os.path.basename(r) for r in product_repos()]
    libs = [d for d in editable_products if d in available_products and d in product_dependencies and tox.workdir in product_dependencies[d]]

    already_editable = [d for d in editable_products if d in product_dependencies and tox.workdir not in product_dependencies[d]]
    for lib in already_editable:
      log.info('%s is already installed in editable mode.', lib)

    not_dependent = [d for d in editable_products if d not in product_dependencies]
    for lib in not_dependent:
      log.debug('%s is not currently installed (not a dependency) and will be ignored.', lib)

    not_available = [d for d in editable_products if d not in not_dependent and d not in available_products]
    for lib in not_available:
      log.info('%s is a dependency but not checked out in workspace, and so can not be installed in editable mode.', lib)

    pip = tox.bindir(env, 'pip')

    for lib in libs:
      if not self.silent or self.debug:
        log.info('%s: Installing %s in editable mode' % (env, lib))

      with log_exception('An error occurred when installing %s in editable mode' % lib):
        run([pip, 'uninstall', lib, '-y'], raises=False, silent=not self.debug)

        lib_path = product_path(lib)
        if os.path.exists(os.path.join(lib_path, lib)):
          lib_path = os.path.join(lib_path, lib)
        run([pip, 'install', '--editable', lib_path], silent=not self.debug)

  def product_depends_on(self, path, name):
    for req_file in config.bump.requirement_files.split():
      req_path = os.path.join(path, req_file)
      if os.path.exists(req_path):
        with open(req_path) as fp:
          try:
            reqs = pkg_resources.parse_requirements(fp.read())
            if name in [r.project_name for r in reqs]:
              return True
          except:
            pass
    return False


def test_repo(repo, test_args, test_class):
  name = product_name(repo)

  branch = current_branch(repo)
  on_branch = '#' + branch if branch != 'master' else ''
  log.info('Testing %s %s', name, on_branch)

  return name, test_class(repo=repo, **dict(test_args)).run()



import logging
import sys

from workspace.commands import AbstractCommand
from workspace.scm import is_git_repo, repo_check

log = logging.getLogger(__name__)


class Review(AbstractCommand):
  """
    Create or update a ReviewBoard.

    Note: This is an abtract command, though may change in the future.
    To use review command, please implement all the abstract methods using the arguments.

    :param str files: [svn only] List of files to use for RB.
    :param int rb_id: RB id to update. For git, this is auto looked up in .git/config
    :param str description: [svn only] Description for RB
    :param bool push: Wait for "Ship It" for RB and push the change. This implies --publish.
    :param int test: Update testing done section with results from tests. Repeat twice to test dependents too (-tt)
  """
  alias = 'rb'

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return ([
        cls.make_args('files', nargs='*', help=docs['files']),
        cls.make_args('-r', '--rb-id', type=int, help=docs['rb_id']),
        cls.make_args('-m', '--description', help=docs['description']),
        cls.make_args('-P', '--publish', action='store_true', help='Publish the RB'),
      ], [
        cls.make_args('-t', '--test', action='count', help=docs['test']),
        cls.make_args('-p', '--push', action='store_true', help=docs['push'])
      ])

  def run(self):

    repo_check()

    if not self.skip_prereview:
      self.prereview_test()

    tests = None

    if self.test:

      if isinstance(self.test, int):
        log.info('Running tests')
        tests = self.commander.run('test', return_output=True, test_dependents=self.test > 1)
      else:
        tests = self.test

      success, tests = self.commander.command('test').summarize(tests)

      if not success:
        sys.exit(1)

      if not isinstance(tests, list):
        tests = [tests]

      self.tests = '\n'.join(filter(lambda t: 'No tests' not in t, tests))

    if not self.rb_id and is_git_repo():
      self.rb_id = self.id_for_branch()

    if self.rb_id:
      self.update()

    else:
      if not self.reviewer_groups:
        self.reviewer_groups = set()
      if not self.reviewers:
        self.reviewers = set()
      g, r = self.reviewers_for_product()
      if g:
        self.reviewer_groups.update(g)
      if r:
        self.reviewers.update(r)
      self.create()

    if self.push:
      if is_git_repo():
        self.commander.run('wait', push=True, rb_id=self.rb_id, in_background=True)
      else:
        log.error('--push is not supported for svn yet.')

  def prereview_test(self):
    """
      Run local tests before creating the RB, such as doing a style check.
      It should run with silent=2 (silent but show output on error).
    """
    if self.commander.command('test').supports_style_check():
      log.info('Running style check')
      self.commander.run('test', env_or_file=['style'], silent=2)

  @classmethod
  def id_for_branch(cls, branch=None):
    """ Returns the RB ID for the current or given branch. """
    raise NotImplementedError('Not implemented. Please implement Review.id_for_branch() in a subclass')

  @classmethod
  def reviewers_for_product(cls, product=None):
    """ Tuple of set(reviewer_groups), set(reviewers) for the current or given product. """
    raise NotImplementedError('Not implemented. Please implement Review.reviewers_for_product() in a subclass')

  def update(self):
    """ Updates the given self.rb_id based on args. Raises on error """
    raise NotImplementedError('Not implemented. Please implement Review.update() in a subclass')

  def create(self):
    """ Creates a RB based on args. Raises on error """
    raise NotImplementedError('Not implemented. Please implement Review.create() in a subclass')



from workspace.commands.commit import Commit


def test_branch_for_msg():
  assert 'fix' == Commit()._branch_for_msg('Fix a big bad bug', 1)
  assert 'fix-a-big' == Commit()._branch_for_msg('Fix a big bad bug', 2)
  assert 'fix-a-big-bad' == Commit()._branch_for_msg('Fix a big bad bug', 3)

  assert 'fix-a-big-bad-bug' == Commit()._branch_for_msg('Fix a big bad bug', 3, ['fix-a-big-bad'])

  assert 'fix-flake8-errors' == Commit()._branch_for_msg('Fix flake8 errors', 3)

  assert 'use-or-or-or-blah-23242-last' == Commit()._branch_for_msg('Use / or | or . or blah_23242_last ignored(&(&*2324', 3)

  assert 'remove-readme-rst' == Commit()._branch_for_msg('Remove README.rst symlink', 3)

  assert 'publish-version-0-4-2' == Commit()._branch_for_msg('Publish version 0.4.2', 3)

  assert 'fix-test' == Commit()._branch_for_msg('Fix test', 3)
  assert 'fix-test' == Commit()._branch_for_msg('Fix test to', 3)
  assert 'fix-test-to-work' == Commit()._branch_for_msg('Fix test to work', 3)



import shlex

import pytest

from workspace.controller import Commander


@pytest.fixture()
def run(monkeypatch):
  def _run(cmd):
    monkeypatch.setattr('sys.argv', shlex.split('wst ' + cmd))
    return Commander().run()
  return _run



#!/usr/bin/env python2.6

import os
import setuptools


def find_files(path):
  return [os.path.join(path, f) for f in os.listdir(path)]


setuptools.setup(
  name='workspace-tools',
  version='1.0.11',

  author='Max Zheng',
  author_email='maxzheng.os @t gmail.com',

  description='Tools to simplify working with multiple product repositories with SCM / development tools abstraction.',
  long_description=open('README.rst').read(),

  changelog_url='https://raw.githubusercontent.com/maxzheng/workspace-tools/master/docs/CHANGELOG.rst',
  url='https://github.com/maxzheng/workspace-tools',

  entry_points={
    'console_scripts': [
      'wst = workspace.controller:Commander.main',
    ],
  },

  install_requires=open('requirements.txt').read(),

  license='MIT',

  package_dir={'': 'src'},
  packages=setuptools.find_packages('src'),
  include_package_data=True,

  setup_requires=['setuptools-git'],

  classifiers=[
    'Development Status :: 5 - Production/Stable',

    'Intended Audience :: Developers',
    'Topic :: Software Development',

    'License :: OSI Approved :: MIT License',

    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
  ],

  keywords='workspace multiple repositories git svn scm abstraction development tools',
)



import os
import pytest

from mock import patch, Mock

from bumper.utils import PyPI

from workspace.scm import stat_repo, all_branches, commit_logs

from test_stubs import temp_dir, temp_git_repo, temp_remote_git_repo


@pytest.mark.parametrize('command,exception', [('diff', None), ('log', SystemExit), ('status', None)])
def test_sanity(run, command, exception):
  with temp_dir():
    if exception:
      with pytest.raises(exception):
        run(command)
    else:
      run(command)

  with temp_git_repo():
    run(command)


def test_bump(run, monkeypatch):
  monkeypatch.setattr('workspace.commands.review.Review.reviewers_for_product', Mock(return_value=(None, None)))
  monkeypatch.setattr('workspace.commands.review.Review.id_for_branch', Mock(return_value=None))
  monkeypatch.setattr('workspace.commands.review.Review.create', Mock())

  with temp_dir():
    with pytest.raises(SystemExit):
      run('bump')

  with temp_remote_git_repo():
    # No requirements.txt
    if os.path.exists('requirements.txt'):
      os.unlink('requirements.txt')
    with pytest.raises(SystemExit):
      run('bump')

    # All requirements are up to date
    with open('requirements.txt', 'w') as fp:
      fp.write('localconfig\nrequests')
    assert ({}, None, []) == run('bump')

    # All requirements are outdated
    with open('requirements.txt', 'w') as fp:
      fp.write('# Comment for localconfig\nlocalconfig==0.0.1\n# Comment for requests\nrequests<0.1')
    msgs, commit_msg, bumps = run('bump')
    file, msg = msgs.items()[0]
    version = PyPI.latest_package_version('localconfig')
    assert 'requirements.txt' == file

    expected_msg = 'Require localconfig==%s' % version
    assert expected_msg == msg[:len(expected_msg)]
    assert expected_msg == commit_msg[:len(expected_msg)]

    with open('requirements.txt') as fp:
      requirements = fp.read()
      assert '# Comment for localconfig\nlocalconfig==%s\n# Comment for requests\nrequests<0.1\n' % version == requirements


def test_cleanrun(run):
  with temp_dir():
    run('clean')


def test_commit(run):
  with temp_dir():
    with pytest.raises(SystemExit):
      run('commit')

  with temp_git_repo():
    with pytest.raises(SystemExit):
      run('commit "no files to commit"')

    with open('new_file', 'w') as fp:
      fp.write('Hello World')
    assert 'new_file' in stat_repo(return_output=True)

    run('commit "Add new file" --branch master')

    assert 'working directory clean' in stat_repo(return_output=True)
    assert 'Hello World' == open('new_file').read()

    with open('new_file', 'w') as fp:
      fp.write('New World')

    run('commit "Update file" --branch updated')

    assert ['updated', 'master'] == all_branches()

    run('commit --move moved')

    assert ['updated', 'master', 'moved'] == all_branches()

    run('commit --discard')

    assert ['master', 'moved'] == all_branches()

    logs = commit_logs()
    assert 'new file' in logs
    assert 1 == len(filter(None, logs.split('commit')))


def test_test(run):
  with temp_dir():
    with pytest.raises(SystemExit):
      run('test')

  if 'PYTESTARGS' in os.environ:
    del os.environ['PYTESTARGS']

  with temp_git_repo():
    with pytest.raises(SystemExit):
      run('test')
    run('setup --product')

    pass_test = 'def test_pass():\n  assert True'
    fail_test = 'def test_fail():\n  assert False'

    with open('test/test_pass.py', 'w') as fp:
      fp.write(pass_test)
    commands = run('test')
    assert 'py27' in commands
    assert 'tox' in commands['py27']

    with open('test/test_fail.py', 'w') as fp:
      fp.write(fail_test + '\n' + pass_test)
    with pytest.raises(SystemExit):
      run('test')

    commands = run('test test/test_pass.py')
    assert 'py27' in commands
    assert 'py.test' in commands['py27']

    os.utime('requirements.txt', None)
    commands = run('test -k test_pass')
    assert 'py27' in commands
    assert 'tox' in commands['py27']

    with pytest.raises(SystemExit):
      run('test style')
    with open('test/test_fail.py', 'w') as fp:
      fp.write(fail_test + '\n\n\n' + pass_test)
    assert 'style' in run('test style')

    os.unlink('test/test_fail.py')
    assert 'coverage' in run('test coverage')
    assert os.path.exists('coverage.xml')
    assert os.path.exists('htmlcov/index.html')


@patch('workspace.commands.push.push_repo')
def test_push(push_repo, run):
  with temp_dir():
    with pytest.raises(SystemExit):
      run('push')

  with temp_remote_git_repo():
    run('push')

    with open('new_file', 'w') as fp:
      fp.write('Hello World')
    assert 'new_file' in stat_repo(return_output=True)

    run('commit "Add new file" --branch=new_file')

    run('push')

    assert ['master'] == all_branches()
    assert "ahead of 'origin/master' by 1 commit." in stat_repo(return_output=True)



import os
import pytest

from test_stubs import temp_dir


def test_checkout_with_alias(run):
  with temp_dir():
    with pytest.raises(SystemExit):
      run('checkout foobazbar-no-such-repo')

    run('checkout mzheng-repos')
    assert os.path.exists('workspace-tools')

    run('checkout clicast')
    assert os.path.exists('clicast')


def test_checkout_with_http_git(run):
  with temp_dir():
    run('checkout https://github.com/maxzheng/clicast.git')
    assert os.path.exists('clicast/README.rst')


def test_checkout_with_git(run):
  with temp_dir():
    run('checkout git@github.com:maxzheng/clicast.git')
    assert os.path.exists('clicast/README.rst')


def test_checkout_with_svn(run):
  with temp_dir():
    run('checkout https://github.com/maxzheng/localconfig/trunk')
    assert os.path.exists('localconfig/README.rst')


def test_checkout_with_multiple_repos(run):
  with temp_dir():
    run('checkout https://github.com/maxzheng/localconfig.git https://github.com/maxzheng/remoteconfig.git')
    assert os.path.exists('localconfig/README.rst')
    assert os.path.exists('remoteconfig/README.rst')



import logging
import sys

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.scm import is_git_repo, checkout_branch, update_repo, repos, product_name, current_branch,\
    update_branch
from workspace.utils import parallel_call

log = logging.getLogger(__name__)


class Update(AbstractCommand):
  """
  Update current product or all products in workspace

  :param list products: When updating all products, filter by these products or product groups
  """
  alias = 'up'

  def __init__(self, *args, **kwargs):
    kwargs.setdefault('raises', True)
    super(Update, self).__init__(*args, **kwargs)

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [cls.make_args('products', nargs='*', help=docs['products'])]

  def run(self):

    if self.products:
      self.products = expand_product_groups(self.products)

    select_repos = [repo for repo in repos() if not self.products or self.products and product_name(repo) in self.products]

    if not select_repos:
      log.info('No product found')

    elif len(select_repos) == 1:
      _update_repo(select_repos[0], self.raises)

    else:
      if not all(parallel_call(_update_repo, select_repos).values()):
        sys.exit(1)


def _update_repo(repo, raises=False):
  name = product_name(repo)

  log.info('Updating %s', name)

  try:
    branch = None
    if is_git_repo(repo):
      branch = current_branch(repo)
      if branch != 'master':
        checkout_branch('master', repo)

    update_repo(repo)

    if branch and branch != 'master':
      log.info('Rebasing %s', branch)
      checkout_branch(branch, repo)
      update_branch(repo)

    return True
  except Exception as e:
    if raises:
      raise
    else:
      log.error('%s: %s', name, e)
      return False



from glob import glob
import logging
import os
import re
import sys

from localconfig import LocalConfig
from workspace.config import product_groups

from workspace.commands import AbstractCommand
from workspace.commands.helpers import expand_product_groups
from workspace.scm import is_repo, repo_check, product_name, repo_path, product_path
from workspace.utils import run


log = logging.getLogger(__name__)

BASHRC_FILE = "~/.bashrc"
WSTRC_FILE = "~/.wstrc"

WS_SETUP_START = '# Added by "workspace setup" (do not remove comments before / after function)'
WS_SETUP_END = "# workspace setup - end"

WS_FUNCTION_TEMPLATE = """\
alias _wst=%s

function ws() {
  if [ $# -gt 0 ]; then
    _wst "$@";
  else
    cd %s
    ls
  fi
}

function activate() {
  if [ -e activate ]; then
    source activate
  elif [ -e ${PWD##*/}/activate ]; then
    source ${PWD##*/}/activate
  else
    source .tox/${PWD##*/}/bin/activate
  fi
}

function open_files_from_last_command() {
  if [ "$1" ]; then
    last_command="1 ag '$@'"
  else
    last_command=`history 100 | grep  -E "^\s+[0-9]+\s+(ag|ack|grep|find|which) " | tail -1`
  fi

  if [ -z "$last_command" ]; then
    echo No ag, ack, grep, find, or which command found in last 100 commands.
    return
  fi

  declare -a "parts=($last_command)"
  command=${parts[1]}

  if [[ "$command" = "ag" || "$command" = "ack" || "$command" = "grep" ]]; then
    full_command=${parts[@]:1}
    pattern=+/${parts[2]}

    raw_parts=(${last_command// / })  # Need the quote retained to sub properly
    last_part=${raw_parts[@]:(-1)}

    if [[ $last_part != "-l" ]]; then
        sub_expr=" $last_part= $last_part -l"
    else
        sub_expr=
    fi

    if [ "$1" ]; then
      files=`ag -l "$@"`
    elif [ -z "$sub_expr" ]; then
      files=`fc -s $command`
    else
      files=`fc -s "$sub_expr" $command`
    fi

    if [ -z "$files" ]; then
      echo No files found from output
    else
      vim -p $files "$pattern" --cmd "set ignorecase smartcase"
    fi

  else
    files=`fc -s $command`

    if [ -z "$files" ]; then
      echo No files found from output
    else
      vim -p $files
    fi

  fi
}
"""
COMMAND_FUNCTION_TEMPLATE = 'function %s() { _wst %s "$@"; }\n'
COMMAND_ALIAS_TEMPLATE = 'alias %s=%s\n'
COMMANDS = {
  'a': "'activate'",
  'd': "'deactivate'",
  'tv': "'open_files_from_last_command'  # from ag/ack/grep/find/which [t]o [v]im",

  'co': 'checkout',
  'ci': 'commit',
  'di': '_diff',
  'rb': 'review',
  'st': 'status',
  'up': 'update',
  'w8': '_wait',

  '_bu': 'bump',
  '_cl': 'clean',
  '_lo': 'log',
  '_pu': 'push',
  '_pb': 'publish',
  '_te': 'test',
}
AUTO_COMPLETE_TEMPLATE = """
function _branch_file_completer() {
  local cur=${COMP_WORDS[COMP_CWORD]}

  if git status &> /dev/null; then
    branches=`git branch`
  else
    branches=
  fi

  if [ ! -z "$branches" ]; then
    COMPREPLY=( $( compgen -W "$branches" -- $cur ) )
  fi
}
function _env_file_completer() {
  local cur=${COMP_WORDS[COMP_CWORD]}

  if ls tox*.ini &>/dev/null || ls .tox*.ini &>/dev/null; then
    envs=`grep -h '^\[testenv:' .tox*.ini tox*.ini 2>/dev/null | sed -E 's/^\[testenv:(.+)]/\\1/' | grep -vE '^(py|pydev)$'`
    COMPREPLY=( $( compgen -W "$envs" -- $cur ) )
  fi
}

complete -o default -F _branch_file_completer co
complete -o default -F _branch_file_completer checkout
complete -o default -F _env_file_completer test
complete -F _branch_file_completer push

complete -o default log
complete -o default di
"""
TOX_INI_FILE = 'tox.ini'
TOX_INI_TMPL = """\
[tox]
envlist = py27

[testenv]
commands =
    py.test {env:PYTESTARGS:}
install_command = pip install -U {packages}
recreate = False
skipsdist = True
usedevelop = True
setenv =
    PIP_PROCESS_DEPENDENCY_LINKS=1
    PIP_DEFAULT_TIMEOUT=60
    ARCHFLAGS=-Wno-error=unused-command-line-argument-hard-error-in-future
basepython = python

[testenv:py]
deps =
    pytest
    pytest-xdist

[testenv:pydev]
deps =
    {[testenv:py]deps}
    {[testenv:style]deps}
    pytest-cov
    sphinx!=1.2b2

[testenv:py27]
deps = {[testenv:pydev]deps}
envdir = {toxworkdir}/%s
basepython = python2.7

[testenv:style]
commands =
    flake8 --config tox.ini src test
deps =
    flake8

[testenv:coverage]
commands =
    py.test {env:PYTESTARGS:} --cov=src --cov-report=xml --cov-report=html --cov-report=term test
deps =
    {[testenv:py]deps}
    pytest-cov

[flake8]
ignore = E111,E121,W292,E123,E226
max-line-length = 160

[pytest]
addopts = -n 4

[wst]
template_version = 1
"""
SETUP_PY_TMPL = """\
#!/usr/bin/env python

import os
import setuptools


setuptools.setup(
  name='%s',
  version='0.0.1',

  author='<PLACEHOLDER>',
  author_email='<PLACEHOLDER>',

  description='<PLACEHOLDER>',
  long_description=open('%s').read(),

  url='<PLACEHOLDER>',

#  entry_points={
#    'console_scripts': [
#      'script_name = package.module:entry_callable',
#    ],
#  },

  install_requires=open('%s').read(),

  license='MIT',

  package_dir={'': 'src'},
  packages=setuptools.find_packages('src'),
  include_package_data=True,

  setup_requires=['setuptools-git'],

#  scripts=['bin/cast-example'],

  classifiers=[
    'Development Status :: 5 - Production/Stable',

    'Intended Audience :: Developers',
    'Topic :: Software Development :: <PLACEHOLDER SUB-TOPIC>',

    'License :: OSI Approved :: MIT License',

    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
  ],

  keywords='<KEYWORDS>',
)
"""
README_TMPL = """\
%s
===========

<PLACEHOLDER DESCRIPTION>
"""


class Setup(AbstractCommand):
  """
    Sets up workspace or product environment.

    :param str product_group: Setup product group by checking them out, developing them, and running any setup scripts and
                              exports as defined by setup.cfg in each product.
    :param bool product: Initialize product by setting up tox with py27, style, and coverage test environments.
                         Also create setup.py, README.rst, and src / test directories if they don't exist.
    :param bool commands: Add convenience bash function for certain commands, such as checkout to run
                          "workspace checkout", or "ws" bash function that goes to your workspace directory
                          when no argument is passed in, otherwise runs wst command.
    :param bool commands_with_aliases: Same as --commands plus add shortcut aliases, like "co" for checkout.
                                       This is for those developers that want to get as much done with the least
                                       key strokes - true efficienist! ;)
    :param bool uninstall: Uninstall all functions/aliases.
  """
  #: Dict for additional commands to setup
  additional_commands = None

  @classmethod
  def arguments(cls):
    _, docs = cls.docs()
    return [
      cls.make_args('product_group', nargs='?', help=docs['product_group']),
      cls.make_args('--product', action='store_true', help=docs['product']),
      cls.make_args('--commands', action='store_true', help=docs['commands']),
      cls.make_args('-a', '--commands-with-aliases', action='store_true', help=docs['commands_with_aliases']),
      cls.make_args('--uninstall', action='store_true', help=docs['uninstall'])
    ]

  def run(self):
    num_options = len(filter(None, [self.product_group, self.product, self.commands, self.commands_with_aliases, self.uninstall]))
    if num_options > 1:
      log.error('Only one setup option can be selected at a time.')
      sys.exit(1)

    elif not num_options:
      log.error('At least one option must be selected. See -h for options.')
      sys.exit(1)

    if self.product_group:
      self.setup_product_group()
    elif self.product:
      self.setup_product()
    else:
      self.setup_workspace()

  def setup_product_group(self):
    if is_repo():
      log.error('This should be run from your workspace directory and not within a product repo')
      sys.exit(1)

    if self.product_group not in product_groups():
      log.error('Product group "%s" is not defined in workspace.cfg', self.product_group)
      sys.exit(1)

    log.info('Setting up %s products', self.product_group)

    # Checkout product
    self.commander.run('checkout', target=[self.product_group])

    # Develop the environment
    current_dir = os.getcwd()

    for product in expand_product_groups([self.product_group]):
      log.info('Developing environment for %s', product)
      try:
        repo = product_path(product)
        os.chdir(repo)
        self.commander.run('test', redevelop=True, install_only=True)

      except Exception as e:
        log.error('Error occurred when developing %s: %s', product, e)

      finally:
        os.chdir(current_dir)

    # Process setup.cfg
    exports = {}
    products = expand_product_groups([self.product_group])
    for product in products:
      repo = product_path(product)

      if os.path.join(repo, product, 'setup.cfg'):
        setup_cfg = os.path.join(repo, product, 'setup.cfg')
      else:
        setup_cfg = os.path.join(repo, 'setup.cfg')

      if os.path.exists(setup_cfg):
        setup = LocalConfig(setup_cfg)
        setup._parser.optionxform = str
        if 'scripts' in setup or 'exports' in setup:
          log.info('Processing scripts/exports in setup.cfg for %s', product)

          if 'scripts' in setup:
            cwd = os.getcwd()
            try:
              os.chdir(repo)
              for name, script in setup.scripts:
                log.info('Running %s', name)
                run(['bash', '-c', '; '.join(filter(None, script.split('\n')))])
            except Exception as e:
              log.error('Error occurred running script: %s', e)
            finally:
              os.chdir(cwd)

          if 'exports' in setup:
            for name, value in setup.exports:
              exports[name] = value

    if exports:
      activate_group = 'activate_%s' % self.product_group
      with open(activate_group, 'w') as fp:
        for name in sorted(exports):
          fp.write('export %s=%s\n' % (name, exports[name]))
        fp.write('\n')

        fp.write('if [[ $PS1 != *{%s}* ]]; then\n' % self.product_group)
        fp.write('  export PS1="{%s}$PS1"\n' % self.product_group)
        fp.write('fi\n\n')

        fp.write('deactivate_%s() {\n' % self.product_group)
        for name in sorted(exports):
          fp.write('  unset %s\n' % name)
        fp.write('\n')
        fp.write('  export PS1=${PS1/{%s\}/}\n' % self.product_group)
        fp.write('  unset deactivate_%s\n' % self.product_group)
        fp.write('}\n')
      log.info('Created ./%s. To activate, run: source %s. To deactivate, run: deactivate_%s', activate_group, activate_group, self.product_group)

  def setup_product(self):
    repo_check()

    name = product_name(repo_path())
    placeholder_info = '- please update <PLACEHOLDER> with appropriate value'

    tox_ini = TOX_INI_TMPL % name
    tox_ini_file = os.path.join(repo_path(), TOX_INI_FILE)
    tox_change_word = 'Updated' if os.path.exists(tox_ini_file) else 'Created'
    with open(tox_ini_file, 'w') as fp:
      fp.write(tox_ini)

    log.info('%s %s', tox_change_word, self._relative_path(tox_ini_file))

    readme_files = glob(os.path.join(repo_path(), 'README*'))
    if readme_files:
      readme_file = readme_files[0]
    else:
      readme_file = os.path.join(repo_path(), 'README.rst')
      with open(readme_file, 'w') as fp:
        fp.write(README_TMPL % name)
      log.info('Created %s %s', self._relative_path(readme_file), placeholder_info)

    setup_py_file = os.path.join(repo_path(), 'setup.py')
    if not os.path.exists(setup_py_file):
      requirements_file = os.path.join(repo_path(), 'requirements.txt')
      if not os.path.exists(requirements_file):
        with open(requirements_file, 'w') as fp:
          pass
        log.info('Created %s', self._relative_path(requirements_file))

      readme_name = os.path.basename(readme_file)
      requirements_name = os.path.basename(requirements_file)

      with open(setup_py_file, 'w') as fp:
        fp.write(SETUP_PY_TMPL % (name, readme_name, requirements_name))

      log.info('Created %s %s', self._relative_path(setup_py_file), placeholder_info)

    src_dir = os.path.join(repo_path(), 'src')
    if not os.path.exists(src_dir):
      package_dir = os.path.join(src_dir, re.sub('[^A-Za-z]', '', name))
      os.makedirs(package_dir)
      init_file = os.path.join(package_dir, '__init__.py')
      open(init_file, 'w').close()
      log.info('Created %s', self._relative_path(init_file))

    test_dir = os.path.join(repo_path(), 'test')
    if not os.path.exists(test_dir):
      os.makedirs(test_dir)
      test_file = os.path.join(test_dir, 'test_%s.py' % re.sub('[^A-Za-z]', '_', name))
      with open(test_file, 'w') as fp:
        fp.write('# Placeholder for tests')
      log.info('Created %s', self._relative_path(test_file))

  def setup_workspace(self):
    bashrc_content = None
    bashrc_path = os.path.expanduser(BASHRC_FILE)
    wstrc_path = os.path.expanduser(WSTRC_FILE)

    bashrc_script = []

    if os.path.exists(bashrc_path):
      with open(bashrc_path) as fh:
        bashrc_content = fh.read()

      skip = False
      for line in bashrc_content.split('\n'):
        if line in (WS_SETUP_START, WS_SETUP_END):
          skip = not skip
          continue
        if not skip and WSTRC_FILE not in line:
          bashrc_script.append(line)

      bashrc_script = '\n'.join(bashrc_script).strip().split('\n')  # could be better

    repo_path = is_repo()
    if repo_path:
      workspace_dir = os.path.dirname(repo_path).replace(os.path.expanduser('~'), '~')
    else:
      workspace_dir = os.getcwd().replace(os.path.expanduser('~'), '~')

    with open(bashrc_path, 'w') as fh:
      if bashrc_script:
        fh.write('\n'.join(bashrc_script) + '\n\n')

      if self.uninstall:
        if os.path.exists(wstrc_path):
          os.unlink(wstrc_path)
        log.info('Removed %s and its sourcing reference from %s', WSTRC_FILE, BASHRC_FILE)
        log.info('Please restart your bash session for the change to take effect')
        return

      fh.write('source %s\n' % WSTRC_FILE)

    with open(wstrc_path, 'w') as fh:
      fh.write(WS_FUNCTION_TEMPLATE % (os.path.realpath(sys.argv[0]), workspace_dir))
      log.info('Added "ws" bash function with workspace directory set to %s', workspace_dir)

      if self.additional_commands:
        COMMANDS.update(self.additional_commands)

      def special(c):
        return c.startswith("'") or c.startswith('"') or c.startswith(' ')

      if self.commands or self.commands_with_aliases:
        functions = sorted([f for f in COMMANDS.values() if not special(f)])
        fh.write('\n')
        for func in functions:
          fh.write(COMMAND_FUNCTION_TEMPLATE % (func, func.lstrip('_')))
        log.info('Added bash functions: %s', ', '.join([f for f in functions if not f.startswith('_')]))

      if self.commands_with_aliases:
        fh.write('\n')
        aliases = [item for item in sorted(COMMANDS.items(), key=lambda x: x[1].lstrip('_')) if not item[0].startswith('_')]
        for alias, command in aliases:
          fh.write(COMMAND_ALIAS_TEMPLATE % (alias, command.lstrip(' ')))
        log.info('Added aliases: %s', ', '.join(["%s=%s" % (a, c.lstrip('_ ')) for a, c in aliases if not special(c)]))
        log.info('Added special aliases: %s', ', '.join(["%s=%s" % (a, c.lstrip('_ ')) for a, c in aliases if special(c)]))

        fh.write(AUTO_COMPLETE_TEMPLATE)

    log.info('To use, run "source %s" or open a new shell.', WSTRC_FILE)

  def _relative_path(self, path):
    if path.startswith(os.getcwd() + os.path.sep):
      path = path[len(os.getcwd())+1:]
    return path



from workspace.commands.helpers import expand_product_groups


def test_expand_product_groups(monkeypatch):
  monkeypatch.setattr('workspace.commands.helpers.product_groups',
                      lambda: {'ws': ['workspace-tools', 'clicast', 'localconfig', 'remoteconfig'],
                               'config': ['localconfig', 'remoteconfig']})

  assert expand_product_groups(['name1', 'name2']) == ['name1', 'name2']
  assert expand_product_groups(['ws', 'name2']) == sorted(['workspace-tools', 'clicast', 'localconfig', 'remoteconfig', 'name2'])
  assert expand_product_groups(['ws', '-localconfig']) == sorted(['workspace-tools', 'clicast', 'remoteconfig'])
  assert expand_product_groups(['ws', '-config']) == sorted(['workspace-tools', 'clicast'])
